/** \file dblib.c
 *
 * \brief general database routines:
 *
 * dblib contains functions which can be used to access CVS files and 
 * extract fields.
 * 
 * It also contains some string and conversion utilities.
 * 
 * \todo clean redevelopment desirable
 */
#ifndef NO_DOXYGEN
#include "prolog.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#ifdef WIN32
# include "direct.h"
#endif
#ifdef UNIX
# include <sys/stat.h>
# include <dirent.h>
# include <sys/vfs.h>
# ifndef FREEBSD
# include <alloca.h>
# endif
#endif
#include "srvcore.h"
#include "srvip.h"
#include "dbtools.h"
#include "hstlib.h"
#include "errors.h"
#include "ttycmd.h"

#define LOINT(k)       ((int)(k))
#define HIINT(k)       ((int)(((SINT32)(k) >> 16) & 0xFFFF))

DBLNKLIST *gFecInfoList = NULL;
#define MAX_STACK 32
static int dbStackPtr = 0;
void *dbStack[ MAX_STACK ];
int dbpush( void * p )
{
  if( dbStackPtr >= MAX_STACK ) return 0;
  dbStack[dbStackPtr++] = p;
  return 1;
}
int dbpop( void ** p )
{
  if( dbStackPtr <= 0 ) return 0;
  *p = dbStack[--dbStackPtr];
  return 1;
}
DBLNKLIST * listCreate( void )
{
  DBLNKLIST *p = (DBLNKLIST *)SystemCalloc(1,sizeof(DBLNKLIST));
  if( p == NULL ) return NULL;

  p->current = NULL;
  p->first = NULL;
  p->last = NULL;
  p->nElements = 0;

  return p;
}
void listDelete( DBLNKLIST * list )
{
  DBLNKLIST_ITEM * item;
  DBLNKLIST_ITEM * previousItem;
  item = list->last;  
  while( item )
  {
    previousItem = item->previous;
    if( item->dataBufferAllocated ) free( item->data ); 
    free( item );
    item = previousItem;
  }
  free( list );
}
void listAdd( DBLNKLIST * list, void * data )
{
  DBLNKLIST_ITEM * li = (DBLNKLIST_ITEM *)SystemCalloc(1,sizeof( DBLNKLIST_ITEM ));

  if (list == NULL)
  {
    feclog("listAdd error: list not yet created");
    return;
  }
  if(li == NULL)
  {
    feclog("listAdd error: malloc failed");
    return;
  }
  li->dataBufferAllocated = 0;
  if( list->last != NULL ) 
  {
    li->previous = list->last;
    list->last->next =  li;		
  }
  else
  {
    list->first = li;
    li->previous = NULL;
  }
  li->next = NULL;
  li->data = data;
  list->current = li;
  list->last = li;
  ++list->nElements;
}
void listAddCopy( DBLNKLIST * list, void * data, int sizeOfData )
{
  DBLNKLIST_ITEM * li = (DBLNKLIST_ITEM *)SystemCalloc(1,sizeof(DBLNKLIST_ITEM));
  void * dataCopy;

  if( li == NULL )
  {
    puts( "listAddCopy error: malloc failed" );
    return;
  }
  if( (dataCopy = SystemCalloc(1,sizeOfData)) == NULL )
  {
    puts( "listAddCopy error: malloc failed" );
    return;
  }
  memcpy( dataCopy, data, sizeOfData );
  li->dataBufferAllocated = 1;
  if( list->last != NULL ) 
  {
    li->previous = list->last;
    list->last->next =  li;		
  }
  else
  {
    list->first = li;
    li->previous = NULL;
  }
  li->next = NULL;
  li->data = dataCopy;
  list->current = li;
  list->last = li;
  ++list->nElements;
}
void listRemoveCurrent( DBLNKLIST * list )
{
  if( list == NULL || list->current == NULL ) return;

  if( list->current->previous != NULL ) list->current->previous->next = list->current->next;
  else list->first = list->current->next;

  if( list->current->next != NULL ) list->current->next->previous = list->current->previous;
  else list->last = list->current->previous;

  if( list->current->dataBufferAllocated ) free( list->current->data );

  free( list->current );
  --list->nElements;
}
void *listGetNext( DBLNKLIST * list )
{
  void * data;
  if( list != NULL && list->current != NULL )
  {
    data = list->current->data;
    list->current = list->current->next;
    return data;
  }
  return NULL;
}
void *listGetPrevious( DBLNKLIST * list )
{
  void * data;
  if( list != NULL && list->current != NULL )
  {
    data = list->current->data;
    list->current = list->current->previous;
    return data;
  }
  return NULL;
}
int listHasNext( DBLNKLIST * list )
{
  if (list == NULL) return 0;
  return ( list->current->next != NULL );
}
int listHasPrevious( DBLNKLIST * list )
{
  if (list == NULL) return 0;
  return ( list->current->previous != NULL );
}
void listGoToFirst( DBLNKLIST * list )
{
  if (list != NULL) list->current = list->first;
}
void listGoToLast( DBLNKLIST * list )
{
  if (list != NULL) list->current = list->last;
}
void freeFecListMemory( DBLNKLIST *fecList )
{
  xmlFIDS *fec;
  xmlXIS *eqm;
  xmlDEVIS *device;
  xmlPRPIS *property;
  xmlSETIS *set;
  xmlAIS *alias;
  xmlADS *almdef;
  xmlHIS *hst;
  xmlAWIS *alm;
  NAME64 *member;

  if( fecList )
  {
    listGoToFirst( fecList );
    while ( fec = listGetNext( fecList ) )
    {
      if( fec->eqms ) 
      {
        listGoToFirst( fec->eqms );			
        while( eqm = listGetNext(  fec->eqms ) )
        {
          if( eqm->devices )
          {
            listGoToFirst( eqm->devices );
            while( device = listGetNext( eqm->devices ) ) SystemFree( device );
            listDelete( eqm->devices );
          }
          if( eqm->properties )
          {
            listGoToFirst( eqm->properties );
            while( property = listGetNext( eqm->properties ) ) 
            {
              if (property->histories)
              {
                listGoToFirst( property->histories );
                while( hst = listGetNext( property->histories ) ) SystemFree( hst );
                listDelete( property->histories );
              }
              if (property->almwatches)
              {
                listGoToFirst( property->almwatches );
                while( alm = listGetNext( property->almwatches ) ) SystemFree( alm );
                listDelete( property->almwatches );
              }
              SystemFree( property );
            }
            listDelete( eqm->properties );
          }
          if( eqm->alarmdefs )
          {
            listGoToFirst( eqm->alarmdefs );
            while( almdef = listGetNext( eqm->alarmdefs ) ) SystemFree( almdef );
            listDelete( eqm->alarmdefs );
          }
          if (eqm->sets)
          {
            listGoToFirst(eqm->sets);
            while( set = listGetNext(eqm->sets) )
            {
              if (set->members)
              {
                listGoToFirst(set->members);
                while( member = listGetNext( set->members ) ) SystemFree( member );
                listDelete(set->members);
              }
              SystemFree(set);
            }
            listDelete(eqm->sets);
          }
          SystemFree( eqm );
        }
        listDelete( fec->eqms );
        if (fec->alias)
        {
          listGoToFirst(fec->alias);
          while( alias = listGetNext(fec->alias) )
          {
            SystemFree(alias);
          }
          listDelete(fec->alias);
        }
        if (fec->sets)
        {
          listGoToFirst(fec->sets);
          while( set = listGetNext(fec->sets) )
          {
            if (set->members)
            {
              listGoToFirst(set->members);
              while( member = listGetNext( set->members ) ) SystemFree( member );
              listDelete(set->members);
            }
            SystemFree(set);
          }
          listDelete(fec->sets);
        }
      }
      SystemFree( fec );
    }
    listDelete( fecList );
  }
}
XML_DBHANDLER *CurrentHandler = NULL;
static void setCurrentHandler( XML_DBHANDLER * xmlHandler )
{
  if( CurrentHandler != NULL ) dbpush( CurrentHandler );
  CurrentHandler = xmlHandler;
}
static void restoreHandler( void )
{
  XML_DBHANDLER * h;
  if( dbpop( (void ** )& h ) ) CurrentHandler = h;
}
static int documentProcessOpeningTag( char *tag )
{
	xmlFIDS * fec;
	
	if( !strcmp( tag, "FEC" ) ) 
	{	/* add new FEC to the list */
		fec = (xmlFIDS *)SystemCalloc(1,sizeof( xmlFIDS )); 
		fec->eqms = listCreate(); 
    fec->sets = listCreate();
    fec->alias = listCreate();
    listAdd( CurrentHandler->currentDataSection, fec );
    setCurrentHandler( &FecHandler );
    CurrentHandler->currentDataSection = fec;
    return 1;
  }
  else if( tag[0] == '?' ) return 1; /* skip header */

  dbglog( "fecProcessOpeningTag( %s ): unknown tag\n", tag );
  return 0;
}
static int documentProcessClosingTag( char *tag )
{
  return 1;
}
static int fecProcessOpeningTag( char *tag )
{
  xmlXIS * eqm;
  xmlSETIS *set;
  xmlAIS *alias;
	
  if( !strcmp( tag, "EQM" ) ) 
  { /* add new EQM to the list */
    eqm = (xmlXIS *)SystemCalloc(1,sizeof( xmlXIS )); 
    eqm->devices = listCreate(); 
    eqm->properties = listCreate();
    eqm->alarmdefs = listCreate();
    eqm->sets = listCreate();
    listAdd( ((xmlFIDS *)CurrentHandler->currentDataSection)->eqms, eqm );
    setCurrentHandler( &EqmHandler );
    CurrentHandler->currentDataSection = eqm;
    return 1;
  }
  if( !strcmp( tag, "NAME" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlFIDS *)(CurrentHandler->currentDataSection))->name;
    return 1;
  }
  if( !strcmp( tag, "PORT_OFFSET" ) ) 
  {
    CurrentHandler->currentDataType = CF_LONG;
    CurrentHandler->currentData = &((xmlFIDS *)(CurrentHandler->currentDataSection))->portOffset;
    return 1;
  }
  if( !strcmp( tag, "HISTORY_HOME" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlFIDS *)(CurrentHandler->currentDataSection))->historyHome;
    return 1;
  }
  if( !strcmp( tag, "NAME_LIST" ) ) 
  { 
    set = (xmlSETIS *)SystemCalloc(1,sizeof( xmlSETIS )); 
    listAdd( ((xmlFIDS *)CurrentHandler->currentDataSection)->sets, set );
    setCurrentHandler( &NameSetHandler );
    CurrentHandler->currentDataSection = set;
    set->members = listCreate(); 
    return 1;
  }
  if( !strcmp( tag, "ALIAS" ) ) 
  { 
    alias = (xmlAIS *)SystemCalloc(1,sizeof( xmlAIS )); 
    listAdd( ((xmlFIDS *)CurrentHandler->currentDataSection)->alias, alias );
    setCurrentHandler( &AliasHandler );
    CurrentHandler->currentDataSection = alias;
    return 1;
  }
  dbglog( "fecProcessOpeningTag( %s ): unknown tag", tag );
  return 0;
}
static int fecProcessClosingTag( char *tag )
{
  if( !strcmp( tag, "FEC" ) ) restoreHandler();
  return 1;
}
static int eqmProcessOpeningTag( char *tag )
{
  xmlDEVIS *device;
  xmlPRPIS *property;
  xmlADS *almdefs;
  xmlSETIS *set;

  if( !strcmp( tag, "ALARM_DEFINITION" ) ) 
  { /* add new alarm definition to the list */
    almdefs = (xmlADS *)SystemCalloc(1,sizeof( xmlADS )); 
    listAdd( ((xmlXIS *)CurrentHandler->currentDataSection)->alarmdefs, almdefs );
    setCurrentHandler( &AlmDefHandler );
    CurrentHandler->currentDataSection = almdefs;
    return 1;
  }
  if( !strcmp( tag, "NAME_LIST" ) ) 
  { /* add a new name list to the list */
    set = (xmlSETIS *)SystemCalloc(1,sizeof( xmlSETIS )); 
    listAdd( ((xmlXIS *)CurrentHandler->currentDataSection)->sets, set );
    setCurrentHandler( &NameSetHandler );
    CurrentHandler->currentDataSection = set;
    set->members = listCreate(); 
    return 1;
  }
  if( !strcmp( tag, "DEVICE" ) ) 
  { /* add new (common) DEVICE to the list */
    device = (xmlDEVIS *)SystemCalloc(1,sizeof( xmlDEVIS )); 
    listAdd( ((xmlXIS *)CurrentHandler->currentDataSection)->devices, device );
    setCurrentHandler( &DeviceHandler );
    CurrentHandler->currentDataSection = device;
    return 1;
  }
  if( !strcmp( tag, "DEVICE_SPACE" ) ) 
  {
    CurrentHandler->currentDataType = CF_LONG;
    CurrentHandler->currentData = &((xmlXIS*)(CurrentHandler->currentDataSection))->deviceSpace;
    return 1;
  }
  if( !strcmp( tag, "PROPERTY" ) ) 
  { /* add new (common) PROPERTY to the list */
    property = (xmlPRPIS *)SystemCalloc(1,sizeof( xmlPRPIS )); 
    property->histories = listCreate();
    property->almwatches = listCreate();
    listAdd( ((xmlXIS *)CurrentHandler->currentDataSection)->properties, property );
    setCurrentHandler( &PropertyHandler );
    CurrentHandler->currentDataSection = property;

    return 1;
  }
  if( !strcmp( tag, "NAME" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlXIS*)(CurrentHandler->currentDataSection))->name;
    return 1;
  }
  if( !strcmp( tag, "SERVER" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlXIS*)(CurrentHandler->currentDataSection))->serverName;
    return 1;
  }
  if( !strcmp( tag, "GROUP" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlXIS*)(CurrentHandler->currentDataSection))->serverGroup;
    return 1;
  }
  if( !strcmp( tag, "GROUP_INDEX" ) ) 
  {
    CurrentHandler->currentDataType = CF_LONG;
    CurrentHandler->currentData = &((xmlXIS*)(CurrentHandler->currentDataSection))->serverGroupIndex;
    return 1;
  }
  if( !strcmp( tag, "CONTEXT" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlXIS*)(CurrentHandler->currentDataSection))->context;
    return 1;
  }
  if( !strcmp( tag, "SUBSYSTEM" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlXIS*)(CurrentHandler->currentDataSection))->subsystem;
    return 1;
  }
  dbglog( "eqmProcessOpeningTag( %s ): unknown tag", tag );
  return 0;
}
static int eqmProcessClosingTag( char *tag )
{
  if( !strcmp( tag, "EQM" ) )  restoreHandler();
  return 1;
}
static int deviceProcessOpeningTag( char *tag )
{
  if( !strcmp( tag, "NAME" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlDEVIS*)(CurrentHandler->currentDataSection))->name;
    return 1;
  }
  if( !strcmp( tag, "PROPERTY_SET" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlDEVIS*)(CurrentHandler->currentDataSection))->propertySet;
    return 1;
  }
  if( !strcmp( tag, "NUMBER" ) ) 
  {
    CurrentHandler->currentDataType = CF_LONG;
    CurrentHandler->currentData = &((xmlDEVIS*)(CurrentHandler->currentDataSection))->number;
    return 1;
  }
  if( !strcmp( tag, "REDIRECTION" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlDEVIS*)(CurrentHandler->currentDataSection))->redirection;
    return 1;
  }
  dbglog( "deviceProcessOpeningTag( %s ): unknown tag", tag );
  return 0;
}
static int deviceProcessClosingTag( char *tag )
{
  if( !strcmp( tag, "DEVICE" ) )  restoreHandler();
  return 1;
}
static int defaultProcessData( char *data )
{
  switch( CurrentHandler->currentDataType )
  {
    case CF_TEXT:
      strcpy( CurrentHandler->currentData, data );
      CurrentHandler->currentDataType = CF_NULL;
      break;
    case CF_LONG:
      *((SINT32 *)(CurrentHandler->currentData)) = atoi( data );
      CurrentHandler->currentDataType = CF_NULL;
      break;
    case CF_FLOAT:
      *((float *)(CurrentHandler->currentData)) = (float)atof( data );
      CurrentHandler->currentDataType = CF_NULL;
      break;
    case CF_DOUBLE:
      *((double*)(CurrentHandler->currentData)) = atof( data );
      CurrentHandler->currentDataType = CF_NULL;
      break;
    case CF_NULL:
      return 1;
      break;
    default:
      dbglog( "defaultProcessData(%s) : unknown data type: %d", data, CurrentHandler->currentDataType ); 
      return 0;
      break;
  }
  return 1;
}
static int almdefProcessOpeningTag( char *tag )
{
  if( !strcmp( tag, "TAG" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlADS *)(CurrentHandler->currentDataSection))->almTag;
    return 1;
  }
  if( !strcmp( tag, "DATA_FORMAT" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlADS *)(CurrentHandler->currentDataSection))->dataFormat;
    return 1;
  }
  if( !strcmp( tag, "ALARM_TEXT" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlADS *)(CurrentHandler->currentDataSection))->almText;
    return 1;
  }
  if( !strcmp( tag, "DEVICE_TEXT" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlADS *)(CurrentHandler->currentDataSection))->devText;
    return 1;
  }
  if( !strcmp( tag, "DATA_TEXT" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlADS *)(CurrentHandler->currentDataSection))->datText;
    return 1;
  }
  if( !strcmp( tag, "URL" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlADS *)(CurrentHandler->currentDataSection))->urlText;
    return 1;
  }
  if( !strcmp( tag, "ALARM_CODE" ) ) 
  {
    CurrentHandler->currentDataType = CF_LONG;
    CurrentHandler->currentData = &((xmlADS *)(CurrentHandler->currentDataSection))->code;
    return 1;
  }
  if( !strcmp( tag, "ALARM_MASK" ) ) 
  {
    CurrentHandler->currentDataType = CF_LONG;
    CurrentHandler->currentData = &((xmlADS *)(CurrentHandler->currentDataSection))->mask;
    return 1;
  }
  if( !strcmp( tag, "SEVERITY" ) ) 
  {
    CurrentHandler->currentDataType = CF_LONG;
    CurrentHandler->currentData = &((xmlADS *)(CurrentHandler->currentDataSection))->severity;
    return 1;
  }
  if( !strcmp( tag, "ALARM_SYSTEM" ) ) 
  {
    CurrentHandler->currentDataType = CF_LONG;
    CurrentHandler->currentData = &((xmlADS *)(CurrentHandler->currentDataSection))->system;
    return 1;
  }
  if( !strcmp( tag, "DATA_SIZE" ) ) 
  {
    CurrentHandler->currentDataType = CF_LONG;
    CurrentHandler->currentData = &((xmlADS *)(CurrentHandler->currentDataSection))->dataSize;
    return 1;
  }
  dbglog( "almdefProcessOpeningTag( %s ): unknown tag", tag );
  return 0;	
}
static int almdefProcessClosingTag( char *tag )
{
  if( !strcmp( tag, "ALARM_DEFINITION" ) )  restoreHandler();
  return 1;
}
static int historyProcessOpeningTag( char *tag )
{
  if( !strcmp( tag, "DEVICE_NAME" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlHIS*)(CurrentHandler->currentDataSection))->hstDevName;
    return 1;
  }
  if( !strcmp( tag, "TOLERANCE" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlHIS*)(CurrentHandler->currentDataSection))->hstTolerance;
    return 1;
  }
  if( !strcmp( tag, "REDIRECTION" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlHIS*)(CurrentHandler->currentDataSection))->hstRemoteServer;
    return 1;
  }
  if( !strcmp( tag, "INDEX" ) ) 
  {
    CurrentHandler->currentDataType = CF_LONG;
    CurrentHandler->currentData = &((xmlHIS*)(CurrentHandler->currentDataSection))->hstIndex;
    return 1;
  }
  if( !strcmp( tag, "HEARTBEAT" ) ) 
  {
    CurrentHandler->currentDataType = CF_LONG;
    CurrentHandler->currentData = &((xmlHIS*)(CurrentHandler->currentDataSection))->hstHeartbeat;
    return 1;
  }
  if( !strcmp( tag, "POLL_RATE" ) ) 
  {
    CurrentHandler->currentDataType = CF_LONG;
    CurrentHandler->currentData = &((xmlHIS*)(CurrentHandler->currentDataSection))->hstPollRate;
    return 1;
  }
  if( !strcmp( tag, "ARCHIVE_RATE" ) ) 
  {
    CurrentHandler->currentDataType = CF_LONG;
    CurrentHandler->currentData = &((xmlHIS*)(CurrentHandler->currentDataSection))->hstArchRate;
    return 1;
  }
  if( !strcmp( tag, "DEPTH_SHORT" ) ) 
  {
    CurrentHandler->currentDataType = CF_LONG;
    CurrentHandler->currentData = &((xmlHIS*)(CurrentHandler->currentDataSection))->hstDepthShort;
    return 1;
  }
  if( !strcmp( tag, "DEPTH_LONG" ) ) 
  {
    CurrentHandler->currentDataType = CF_LONG;
    CurrentHandler->currentData = &((xmlHIS*)(CurrentHandler->currentDataSection))->hstDepthLong;
    return 1;
  }

  dbglog( "historyProcessOpeningTag( %s ): unknown tag", tag );
  return 0;
}
static int alarmProcessOpeningTag( char *tag )
{
  if( !strcmp( tag, "DEVICE_NAME" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlAWIS*)(CurrentHandler->currentDataSection))->almDevName;
    return 1;
  }
  if( !strcmp( tag, "SEVERITY_HIGH" ) ) 
  {
    CurrentHandler->currentDataType = CF_LONG;
    CurrentHandler->currentData = &((xmlAWIS*)(CurrentHandler->currentDataSection))->almSevHigh;
    return 1;
  }
  if( !strcmp( tag, "SEVERITY_LOW" ) ) 
  {
    CurrentHandler->currentDataType = CF_LONG;
    CurrentHandler->currentData = &((xmlAWIS*)(CurrentHandler->currentDataSection))->almSevLow;
    return 1;
  }
  if( !strcmp( tag, "SEVERITY_HIGH_WARN" ) ) 
  {
    CurrentHandler->currentDataType = CF_LONG;
    CurrentHandler->currentData = &((xmlAWIS*)(CurrentHandler->currentDataSection))->almSevHighWarn;
    return 1;
  }
  if( !strcmp( tag, "SEVERITY_LOW_WARN" ) ) 
  {
    CurrentHandler->currentDataType = CF_LONG;
    CurrentHandler->currentData = &((xmlAWIS*)(CurrentHandler->currentDataSection))->almSevLowWarn;
    return 1;
  }
  if( !strcmp( tag, "SYSTEM" ) ) 
  {
    CurrentHandler->currentDataType = CF_LONG;
    CurrentHandler->currentData = &((xmlAWIS*)(CurrentHandler->currentDataSection))->almSystem;
    return 1;
  }
  if( !strcmp( tag, "VALUE_MASK" ) ) 
  {
    CurrentHandler->currentDataType = CF_LONG;
    CurrentHandler->currentData = &((xmlAWIS*)(CurrentHandler->currentDataSection))->almValMask;
    return 1;
  }
  if( !strcmp( tag, "VALUE_NORMAL" ) ) 
  {
    CurrentHandler->currentDataType = CF_LONG;
    CurrentHandler->currentData = &((xmlAWIS*)(CurrentHandler->currentDataSection))->almValNormal;
    return 1;
  }
  if( !strcmp( tag, "COUNT_THRESHOLD" ) ) 
  {
    CurrentHandler->currentDataType = CF_LONG;
    CurrentHandler->currentData = &((xmlAWIS*)(CurrentHandler->currentDataSection))->almCntThreshold;
    return 1;
  }
  if( !strcmp( tag, "VALUE_HIGH" ) ) 
  {
    CurrentHandler->currentDataType = CF_FLOAT;
    CurrentHandler->currentData = &((xmlAWIS*)(CurrentHandler->currentDataSection))->almValHigh;
    return 1;
  }
  if( !strcmp( tag, "VALUE_LOW" ) ) 
  {
    CurrentHandler->currentDataType = CF_FLOAT;
    CurrentHandler->currentData = &((xmlAWIS*)(CurrentHandler->currentDataSection))->almValLow;
    return 1;
  }
  if( !strcmp( tag, "VALUE_HIGH_WARN" ) ) 
  {
    CurrentHandler->currentDataType = CF_FLOAT;
    CurrentHandler->currentData = &((xmlAWIS*)(CurrentHandler->currentDataSection))->almValHighWarn;
    return 1;
  }
  if( !strcmp( tag, "VALUE_LOW_WARN" ) ) 
  {
    CurrentHandler->currentDataType = CF_FLOAT;
    CurrentHandler->currentData = &((xmlAWIS*)(CurrentHandler->currentDataSection))->almValLowWarn;
    return 1;
  }

  dbglog( "alarmProcessOpeningTag( %s ): unknown tag", tag );
  return 0;
}
static int propertyProcessOpeningTag( char *tag )
{
  xmlHIS *history;
  xmlAWIS *alarm;

  if( !strcmp( tag, "NAME" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlPRPIS*)(CurrentHandler->currentDataSection))->name;
    return 1;
  }
  if( !strcmp( tag, "DEVICE_SET" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlPRPIS*)(CurrentHandler->currentDataSection))->deviceSet;
    return 1;
  }
  if( !strcmp( tag, "DESCRIPTION" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlPRPIS*)(CurrentHandler->currentDataSection))->description;
    return 1;
  }
  if( !strcmp( tag, "EGU" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlPRPIS*)(CurrentHandler->currentDataSection))->yEgu;
    return 1;
  }
  if( !strcmp( tag, "XEGU" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlPRPIS*)(CurrentHandler->currentDataSection))->xEgu;
    return 1;
  }
  if( !strcmp( tag, "MAX" ) ) 
  {
    CurrentHandler->currentDataType = CF_FLOAT;
    CurrentHandler->currentData = &((xmlPRPIS*)(CurrentHandler->currentDataSection))->yMax;
    return 1;
  }
  if( !strcmp( tag, "MIN" ) ) 
  {
    CurrentHandler->currentDataType = CF_FLOAT;
    CurrentHandler->currentData = &((xmlPRPIS*)(CurrentHandler->currentDataSection))->yMin;
    return 1;
  }
  if( !strcmp( tag, "XMAX" ) ) 
  {
    CurrentHandler->currentDataType = CF_FLOAT;
    CurrentHandler->currentData = &((xmlPRPIS*)(CurrentHandler->currentDataSection))->xMax;
    return 1;
  }
  if( !strcmp( tag, "XMIN" ) ) 
  {
    CurrentHandler->currentDataType = CF_FLOAT;
    CurrentHandler->currentData = &((xmlPRPIS*)(CurrentHandler->currentDataSection))->xMin;
    return 1;
  }
  if( !strcmp( tag, "REDIRECTION" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlPRPIS*)(CurrentHandler->currentDataSection))->redirection;
    return 1;
  }
  if( !strcmp( tag, "SIZE_IN" ) ) 
  {
    CurrentHandler->currentDataType = CF_LONG;
    CurrentHandler->currentData = &((xmlPRPIS*)(CurrentHandler->currentDataSection))->dataInSize;
    return 1;
  }
  if( !strcmp( tag, "SIZE_OUT" ) ) 
  {
    CurrentHandler->currentDataType = CF_LONG;
    CurrentHandler->currentData = &((xmlPRPIS*)(CurrentHandler->currentDataSection))->dataOutSize;
    return 1;
  }
  if( !strcmp( tag, "DTYPE_IN" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlPRPIS*)(CurrentHandler->currentDataSection))->dataInType;
    return 1;
  }
  if( !strcmp( tag, "DTYPE_OUT" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlPRPIS*)(CurrentHandler->currentDataSection))->dataOutType;
    return 1;
  }
  if( !strcmp( tag, "ACCESS" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlPRPIS*)(CurrentHandler->currentDataSection))->access;
    return 1;
  }
  if( !strcmp( tag, "ID" ) ) 
  {
    CurrentHandler->currentDataType = CF_LONG;
    CurrentHandler->currentData = &((xmlPRPIS*)(CurrentHandler->currentDataSection))->id;
    return 1;
  }
  /* alarm watch entries : */
  if( !strcmp( tag, "ALARM" ) ) 
  {
    alarm = (xmlAWIS *)SystemCalloc(1,sizeof( xmlAWIS )); 
    listAdd( ((xmlPRPIS *)CurrentHandler->currentDataSection)->almwatches, alarm );
    setCurrentHandler( &AlmWatchHandler );
    CurrentHandler->currentDataSection = alarm;
    return 1;
  }
  /* history entries : */
  if( !strcmp( tag, "HISTORY" ) ) 
  { /* add new PROPERTY HISTORY to the list */
    history = (xmlHIS *)SystemCalloc(1,sizeof( xmlHIS )); 
    listAdd( ((xmlPRPIS *)CurrentHandler->currentDataSection)->histories, history );
    setCurrentHandler( &HistoryHandler );
    CurrentHandler->currentDataSection = history;
    return 1;
  }

  dbglog( "propertyProcessOpeningTag( %s ): unknown tag", tag );
  return 0;	
}
static int historyProcessClosingTag( char *tag )
{
  if( !strcmp( tag, "HISTORY" ) )  restoreHandler();
  return 1;
}
static int alarmProcessClosingTag( char *tag )
{
  if( !strcmp( tag, "ALARM" ) )  restoreHandler();
  return 1;
}
static int propertyProcessClosingTag( char *tag )
{
  if( !strcmp( tag, "PROPERTY" ) )  restoreHandler();
  return 1;
}
static int memberProcessOpeningTag( char *tag )
{
  return 1;	
}
static int memberProcessClosingTag( char *tag )
{
  if( !strcmp( tag, "MEMBER" ) )  restoreHandler();
  return 1;
}
static int namesetProcessOpeningTag( char *tag )
{
  NAME64 *member;
  if( !strcmp( tag, "NAME" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlSETIS*)(CurrentHandler->currentDataSection))->name;
    return 1;
  }
  if( !strcmp( tag, "MEMBER" ) ) 
  { 
    CurrentHandler->currentDataType = CF_TEXT;
    member = (NAME64 *)SystemCalloc(1,sizeof( NAME64 )); 
    memcpy(member->name,CurrentHandler->currentDataSection,64);
    listAdd( ((xmlSETIS *)CurrentHandler->currentDataSection)->members, member );
    CurrentHandler->currentData = member;
    return 1;
  }
  dbglog( "namesetProcessOpeningTag( %s ): unknown tag", tag );
  return 0;	
}
static int namesetProcessClosingTag( char *tag )
{
  if( !strcmp( tag, "NAME_LIST" ) )  restoreHandler();
  return 1;
}
static int aliasProcessOpeningTag( char *tag )
{
  if( !strcmp( tag, "NAME" ) ) 
  {
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlAIS*)(CurrentHandler->currentDataSection))->name;
    return 1;
  }
  if( !strcmp( tag, "TARGET" ) ) 
  { 
    CurrentHandler->currentDataType = CF_TEXT;
    CurrentHandler->currentData = ((xmlAIS*)(CurrentHandler->currentDataSection))->target;
    return 1;
  }
  dbglog( "aliasProcessOpeningTag( %s ): unknown tag", tag );
  return 0;	
}
static int aliasProcessClosingTag( char *tag )
{
  if( !strcmp( tag, "ALIAS" ) )  restoreHandler();
  return 1;
}

XML_DBHANDLER DocumentHandler = { documentProcessOpeningTag, documentProcessClosingTag, defaultProcessData, NULL, CF_NULL, NULL };
XML_DBHANDLER FecHandler = { fecProcessOpeningTag, fecProcessClosingTag, defaultProcessData, NULL, CF_NULL, NULL };
XML_DBHANDLER EqmHandler = { eqmProcessOpeningTag, eqmProcessClosingTag, defaultProcessData, NULL, CF_NULL, NULL };
XML_DBHANDLER DeviceHandler = { deviceProcessOpeningTag, deviceProcessClosingTag, defaultProcessData, NULL, CF_NULL, NULL };
XML_DBHANDLER PropertyHandler = { propertyProcessOpeningTag, propertyProcessClosingTag, defaultProcessData, NULL, CF_NULL, NULL };
XML_DBHANDLER AlmDefHandler = { almdefProcessOpeningTag, almdefProcessClosingTag, defaultProcessData, NULL, CF_NULL, NULL };
XML_DBHANDLER NameSetHandler = { namesetProcessOpeningTag, namesetProcessClosingTag, defaultProcessData, NULL, CF_NULL, NULL };
XML_DBHANDLER MemberHandler = { memberProcessOpeningTag, memberProcessClosingTag, defaultProcessData, NULL, CF_NULL, NULL };
XML_DBHANDLER AliasHandler = { aliasProcessOpeningTag, aliasProcessClosingTag, defaultProcessData, NULL, CF_NULL, NULL };
XML_DBHANDLER HistoryHandler = { historyProcessOpeningTag, historyProcessClosingTag, defaultProcessData, NULL, CF_NULL, NULL };
XML_DBHANDLER AlmWatchHandler = { alarmProcessOpeningTag, alarmProcessClosingTag, defaultProcessData, NULL, CF_NULL, NULL };

int populateADS(char *fecname,char *eqmname,DBLNKLIST *fecList,ADSIS **adsis)
{
  xmlFIDS *fec;
  xmlXIS *eqm;
  xmlADS *ads;
  int i, cc = 0, lstsiz = 0;

  if (fecList == NULL || adsis == NULL) ccerr(argument_list_error);
  listGoToFirst(fecList);
  while((fec=listGetNext(fecList)) != NULL) 
  { /* find the fec */
    if (strnicmp(fec->name,fecname,FEC_NAME_SIZE) == 0)
    { /* that's the one */
      listGoToFirst( fec->eqms );
      while((eqm=listGetNext(fec->eqms)) != NULL) 
      { /* find the equipment module */
        if (strncmp(eqm->name,eqmname,EQM_NAME_SIZE)) continue;
        if (eqm->alarmdefs == NULL) continue;
        /* check for common devices */
        listGoToFirst(eqm->alarmdefs);
        while((ads=listGetNext(eqm->alarmdefs)) != NULL) lstsiz++;
        if ((*adsis=(ADSIS *)SystemCalloc(lstsiz,sizeof(ADSIS))) == NULL) ccerr(out_of_server_memory);
        for (i=0,listGoToFirst(eqm->alarmdefs); i<lstsiz && (ads=(xmlADS *)listGetNext(eqm->alarmdefs)) != NULL; i++)
        { 
          strncpy((*adsis)[i].alarmTag,ads->almTag,ALARM_TAG_SIZE);
          strncpy((*adsis)[i].alarmText,ads->almText,ALARM_TEXT_SIZE);
          strncpy((*adsis)[i].deviceText,ads->devText,ALARM_TEXT_SIZE);
          strncpy((*adsis)[i].dataText,ads->datText,ALARM_TEXT_SIZE);
          strncpy((*adsis)[i].url,ads->urlText,ALARM_TEXT_LONGSIZE);
          strncpy((*adsis)[i].alarmDataFormatString,ads->dataFormat,64);
          (*adsis)[i].alarmCode = ads->code;
          (*adsis)[i].alarmMask  = ads->mask;
          (*adsis)[i].alarmDataArraySize = ads->dataSize;
          (*adsis)[i].alarmSeverity = ads->severity;
          (*adsis)[i].alarmSystem = ads->system;
        }
      }
    }
  }
err:
  if (cc) feclog("populate Export Info Database from XML : %s",erlst[cc]);
  return cc ? -cc : lstsiz;
}
int populateAWS(char *fecname,DBLNKLIST *fecList,AWSIS **awsis)
{
  xmlFIDS *fec;
  xmlXIS *eqm;
  xmlPRPIS *prp;
  xmlAWIS *alm;
  int i, cc = 0, lstsiz = 0;

  if (fecList == NULL || awsis == NULL) ccerr(argument_list_error);
  listGoToFirst(fecList);
	while((fec=listGetNext(fecList)) != NULL) 
  { /* find the fec */
    if (strnicmp(fec->name,fecname,FEC_NAME_SIZE) == 0)
    { /* that's the one, now start counting */
      listGoToFirst( fec->eqms );
      while((eqm=listGetNext(fec->eqms)) != NULL) 
      { /* this is a general list over all embedded eqms */
        listGoToFirst(eqm->properties);
        while((prp=listGetNext(eqm->properties)) != NULL) 
        { /* if there are alarm watches associated with the property, count them */
          listGoToFirst(prp->almwatches);
          while ((alm=listGetNext(prp->almwatches)) != NULL) lstsiz++;
        }
      }
      if ((*awsis=(AWSIS *)SystemCalloc(lstsiz,sizeof(AWSIS))) == NULL) ccerr(out_of_server_memory);
      /* now we know how many, so back up */
      i = 0;
      listGoToFirst( fec->eqms );
      while((eqm=listGetNext(fec->eqms)) != NULL) 
      { 
        listGoToFirst(eqm->properties);
        while((prp=listGetNext(eqm->properties)) != NULL) 
        { 
          listGoToFirst(prp->almwatches);
          while ((alm=listGetNext(prp->almwatches)) != NULL)
          {
            if (i >= lstsiz) break;
            strncpy((*awsis)[i].dev,alm->almDevName,DEVICE_NAME_SIZE); /* TODO: allow sets and '*' */
            strncpy((*awsis)[i].alarmDataFormatString,prp->dataOutType,64);
            strncpy((*awsis)[i].prp,prp->name,PROPERTY_NAME_SIZE);
            strncpy((*awsis)[i].eqm,eqm->name,EQM_NAME_SIZE);
            (*awsis)[i].siz = prp->dataOutSize;
            (*awsis)[i].hisev = alm->almSevHigh;
            (*awsis)[i].hiwarnsev = alm->almSevHighWarn;
            (*awsis)[i].losev = alm->almSevLow;
            (*awsis)[i].lowarnsev = alm->almSevLowWarn;
            (*awsis)[i].cntThreshold = alm->almCntThreshold;
            (*awsis)[i].lo = alm->almValLow;
            (*awsis)[i].lowarn = alm->almValLowWarn;
            (*awsis)[i].hi =  alm->almValHigh;
            (*awsis)[i].hiwarn = alm->almValHighWarn;
            (*awsis)[i].asys = alm->almSystem;
            (*awsis)[i].normal = alm->almValNormal;
            (*awsis)[i].mask = alm->almValMask;
            i++;
          }
        }
      }
    }
  }
err:
  if (cc) feclog("populate Export Info Database from XML : %s",erlst[cc]);
  return cc ? -cc : lstsiz;
}
int populateHRS(char *fecname,DBLNKLIST *fecList,HRIS **hris)
{
  xmlFIDS *fec;
  xmlXIS *eqm;
  xmlPRPIS *prp;
  xmlHIS *hst;
  int i, cc = 0, lstsiz = 0;

  if (fecList == NULL || hris == NULL) ccerr(argument_list_error);
  listGoToFirst(fecList);
	while((fec=listGetNext(fecList)) != NULL) 
  { /* find the fec */
    if (strnicmp(fec->name,fecname,FEC_NAME_SIZE) == 0)
    { /* that's the one, now start counting */
      listGoToFirst( fec->eqms );
      while((eqm=listGetNext(fec->eqms)) != NULL) 
      { /* this is a general list over all embedded eqms */
        listGoToFirst(eqm->properties);
        while((prp=listGetNext(eqm->properties)) != NULL) 
        { /* if there are histories associated with the property, count them */
          listGoToFirst(prp->histories);
          while ((hst=listGetNext(prp->histories)) != NULL) lstsiz++;
        }
      }
      if ((*hris=(HRIS *)SystemCalloc(lstsiz,sizeof(HRIS))) == NULL) ccerr(out_of_server_memory);
      /* now we know how many, so back up */
      i = 0;
      listGoToFirst( fec->eqms );
      while((eqm=listGetNext(fec->eqms)) != NULL) 
      { 
        listGoToFirst(eqm->properties);
        while((prp=listGetNext(eqm->properties)) != NULL) 
        { /* if there are histories associated with the property, count them */
          listGoToFirst(prp->histories);
          while ((hst=listGetNext(prp->histories)) != NULL) 
          {
            if (i >= lstsiz) break;
            strncpy((*hris)[i].c.EqmDeviceName,hst->hstDevName,DEVICE_NAME_SIZE); /* TODO: allow sets and '*' */
            strncpy((*hris)[i].c.EqmProperty,prp->name,PROPERTY_NAME_SIZE); 
            strncpy((*hris)[i].c.EqmName,eqm->name,EQM_NAME_SIZE);
            strncpy((*hris)[i].fmt,prp->dataOutType,64);
            strncpy((*hris)[i].redirection,hst->hstRemoteServer,PROPERTY_REDIR_SIZE);
            strncpy((*hris)[i].tolerance,hst->hstTolerance,64);
            (*hris)[i].pollingRate = hst->hstPollRate;
            (*hris)[i].archiveRate = hst->hstArchRate;
            (*hris)[i].depthShort = hst->hstDepthShort;
            (*hris)[i].depthLong = hst->hstDepthLong;
            (*hris)[i].heartbeat = hst->hstHeartbeat;
            (*hris)[i].recordIndex = hst->hstIndex;
            i++;
          }
        }
      }
    }
  }
err:
  if (cc) feclog("populate Export Info Database from XML : %s",erlst[cc]);
  return cc ? -cc : lstsiz;
}
int populateFIDNameList(DBLNKLIST *fecList,char *fecName,char *listName,NAME16 **lst)
{
	xmlFIDS *fec;
  xmlSETIS *set;
  NAME64 *member;
  int i, cc = 0, lstsiz = 0;
  if (fecList == NULL || fecName == NULL || listName == NULL || lst == NULL) ccerr(argument_list_error);
  for (listGoToFirst(fecList);(fec=(xmlFIDS *)listGetNext(fecList)) != NULL;)
  {
    if (strnicmp(fecName,fec->name,FEC_NAME_SIZE) != 0) continue;
    if (fec->sets == NULL) return 0;
    listGoToFirst(fec->sets);
    while ((set=listGetNext(fec->sets)) != NULL)
    {
      if (strnicmp(set->name,listName,SET_NAME_SIZE)) continue;
      /* that's the set !*/
      lstsiz = set->members->nElements;
      listGoToFirst(set->members);
      if ((*lst=(NAME16 *)SystemCalloc(lstsiz,sizeof(NAME16))) == NULL) ccerr(out_of_server_memory);
      for (i=0,listGoToFirst(set->members); i<lstsiz && (member=(NAME64 *)listGetNext(set->members)) != NULL; i++)
      { /* run through eqm->devices for the common devices */
        strncpy((*lst)[i].name,member->name,16);
      }
    }
  }
err:
  if (cc) feclog("populate Allowed Users List from XML : %s",erlst[cc]);
  return lstsiz;
}
int populateFIDS(DBLNKLIST *fecList,FIDS **fids)
{
	xmlFIDS *fec;
  int i, cc = 0, lstsiz = 0;
  if (fecList == NULL || fids == NULL) ccerr(argument_list_error);
  listGoToFirst(fecList);
  while(listGetNext(fecList) != NULL) lstsiz++;
  if ((*fids=(FIDS *)SystemCalloc(lstsiz,sizeof(FIDS))) == NULL) ccerr(out_of_server_memory);
  for (i=0,listGoToFirst(fecList); i<lstsiz && (fec=(xmlFIDS *)listGetNext(fecList)) != NULL; i++)
  {
    strncpy((*fids)[i].fecName,fec->name,FEC_NAME_SIZE);
    (*fids)[i].fecPort = fec->portOffset;
    strncpy((*fids)[i].fecContext,fec->context,CONTEXT_NAME_SIZE);
    strncpy((*fids)[i].fecDescription,fec->description,FEC_DESC_SIZE);
    strncpy((*fids)[i].fecHardware,fec->hardware,FEC_HDW_SIZE);
    strncpy((*fids)[i].fecLocation,fec->location,FEC_LOCATION_SIZE);
    strncpy((*fids)[i].fecSubsystem,fec->subsystem,SUBSYSTEM_NAME_SIZE);
    strncpy((*fids)[i].historyHome,fec->historyHome,HISTORY_HOME_SIZE);
  }
err:
  if (cc) feclog("populate Fec Info Database from XML : %s",erlst[cc]);
  return lstsiz;
}
void xmltoxis(XIS *xis,xmlXIS *eqm,xmlPRPIS *prp)
{
  char xtag[32],ytag[32];
  strncpy(xis->EqmContext,eqm->context,CONTEXT_NAME_SIZE);
  xis->DevSpace = eqm->deviceSpace;
  strncpy(xis->EqmGroup,eqm->serverGroup,EXPORT_NAME_SIZE);
  xis->EqmGroupIdx = eqm->serverGroupIndex;
  strncpy(xis->EqmName,eqm->name,EQM_NAME_SIZE);
  strncpy(xis->ExpName,eqm->serverName,EXPORT_NAME_SIZE);
  strncpy(xis->SubSystem,eqm->subsystem,SUBSYSTEM_NAME_SIZE);
  strncpy(xis->access,prp->access,ACCESS_STR_SIZE);
  strncpy(xis->fmtIn,prp->dataInType,FORMAT_STR_SIZE);
  strncpy(xis->fmtOut,prp->dataOutType,FORMAT_STR_SIZE);
  xis->PrpId = prp->id;
  strncpy(xis->PrpName,prp->name,PROPERTY_NAME_SIZE);
  xis->PrpSizeIn = prp->dataInSize;
  xis->PrpSizeOut = prp->dataOutSize;
  strncpy(xis->redirection,prp->redirection,PROPERTY_REDIR_SIZE);
  if (strchr(prp->description,'[') != NULL && strchr(prp->description,']') != NULL)
  { /* take is as is */
    strncpy(xis->description,prp->description,128);
  }
  else
  { /* prepare it */
    sprintf(ytag,"[%.2g:%.2g %.8s]",prp->yMin,prp->yMax,prp->yEgu);
    xtag[0] = 0;
    if (prp->xMin != 0 || prp->yMin != 0) sprintf(xtag,"[%.2g:%.2g %.8s]",prp->xMin,prp->xMax,prp->xEgu);
    sprintf(xis->description,"%.32s%.32s%.64s",ytag,xtag,prp->description);
  }
}
int populateXIS(char *fecname,char *eqmname,DBLNKLIST *fecList,XIS **xis)
{
  xmlFIDS *fec;
  xmlXIS *eqm;
  xmlPRPIS *prp;
  int i, cc = 0, lstsiz = 0;

  if (fecList == NULL || xis == NULL || eqmname == NULL) ccerr(argument_list_error);
  listGoToFirst(fecList);
  while((fec=listGetNext(fecList)) != NULL) 
  { /* find the fec */
    if (strnicmp(fec->name,fecname,FEC_NAME_SIZE) == 0)
    { /* that's the one */
      listGoToFirst( fec->eqms );
      while((eqm=listGetNext(fec->eqms)) != NULL) 
      { /* find the equipment module */
        if (strncmp(eqm->name,eqmname,EQM_NAME_SIZE)) continue;
        /* check for common properties */
        listGoToFirst(eqm->properties);
        while((prp=listGetNext(eqm->properties)) != NULL) lstsiz++;
        if ((*xis=(XIS *)SystemCalloc(lstsiz,sizeof(XIS))) == NULL) ccerr(out_of_server_memory);
        for (i=0,listGoToFirst(eqm->properties); i<lstsiz && (prp=(xmlPRPIS *)listGetNext(eqm->properties)) != NULL; i++)
        { /* run through eqm->properties for the common properties */
          xmltoxis(&(*xis)[i],eqm,prp);
        }
        ccerr(0); /* done, break out */
      }
    }
  }
err:
  if (cc) feclog("populate Export Info Database from XML : %s",erlst[cc]);
  return cc ? -cc : lstsiz;
}
int populateDEVIS(char *fecname,char *eqmname,DBLNKLIST *fecList,DEVIS **devis)
{
  xmlFIDS *fec;
  xmlXIS *eqm;
  xmlDEVIS *dev;
  int i, cc = 0, lstsiz = 0;

  if (fecList == NULL || devis == NULL) ccerr(argument_list_error);
  listGoToFirst(fecList);
  while((fec=listGetNext(fecList)) != NULL) 
  { /* find the fec */
    if (strnicmp(fec->name,fecname,FEC_NAME_SIZE) == 0)
    { /* that's the one */
      listGoToFirst( fec->eqms );
      while((eqm=listGetNext(fec->eqms)) != NULL) 
      { /* find the equipment module */
        if (strncmp(eqm->name,eqmname,EQM_NAME_SIZE)) continue;
        /* check for common devices */
        listGoToFirst(eqm->devices);
        while((dev=listGetNext(eqm->devices)) != NULL) lstsiz++;
        if ((*devis=(DEVIS *)SystemCalloc(lstsiz,sizeof(DEVIS))) == NULL) ccerr(out_of_server_memory);
        for (i=0,listGoToFirst(eqm->devices); i<lstsiz && (dev=(xmlDEVIS *)listGetNext(eqm->devices)) != NULL; i++)
        { /* run through eqm->devices for the common devices */
          strncpy((*devis)[i].devname,dev->name,DEVICE_NAME_SIZE);
          (*devis)[i].devnum = dev->number;
          strncpy((*devis)[i].devrdr,dev->redirection,DEVICE_REDIR_SIZE);
          /* is there a device specific property list ? */
          strncpy((*devis)[i].prplst,dev->propertySet,SET_NAME_SIZE);;
        }
      }
    }
  }
err:
  if (cc) feclog("populate Export Info Database from XML : %s",erlst[cc]);
  return cc ? -cc : lstsiz;
}
int populatePRPDEVIS(char *fecname,char *eqmname,char *prpname,DBLNKLIST *fecList,DEVIS **devis)
{
  xmlFIDS *fec;
  xmlXIS *eqm;
  xmlPRPIS *prp;
  xmlSETIS *set;
  NAME64 *member;
  int i, cc = 0, lstsiz = 0;

  if (fecList == NULL || devis == NULL) ccerr(argument_list_error);
  listGoToFirst(fecList);
  while((fec=listGetNext(fecList)) != NULL) 
  { /* find the fec */
    if (strnicmp(fec->name,fecname,FEC_NAME_SIZE) == 0)
    { /* that's the one */
      listGoToFirst( fec->eqms );
      while((eqm=listGetNext(fec->eqms)) != NULL) 
      { /* find the equipment module */
        if (strncmp(eqm->name,eqmname,EQM_NAME_SIZE)) continue;
        listGoToFirst(eqm->properties);
        while ((prp=listGetNext(eqm->properties)) != NULL)
        {
          if (strncmp(prp->name,prpname,PROPERTY_NAME_SIZE)) continue;
          /* found the property; does it have a device set ? */
          if (strlen(prp->deviceSet) == 0) ccerr(0);
          /* are there registered sets ? */
          listGoToFirst(eqm->sets);
          while ((set=listGetNext(eqm->sets)) != NULL)
          {
            if (strncmp(set->name,prp->deviceSet,SET_NAME_SIZE)) continue;
            /* that's the set !*/
            lstsiz = set->members->nElements;
            listGoToFirst(set->members);
            if ((*devis=(DEVIS *)SystemCalloc(lstsiz,sizeof(DEVIS))) == NULL) ccerr(out_of_server_memory);
            for (i=0,listGoToFirst(set->members); i<lstsiz && (member=(NAME64 *)listGetNext(set->members)) != NULL; i++)
            { /* run through eqm->devices for the common devices */
              strncpy((*devis)[i].devname,member->name,DEVICE_NAME_SIZE);
              (*devis)[i].devnum = i;
            }
          }
        }
      }
    }
  }
err:
  if (cc) feclog("populate Property-specific Device List from XML : %s",erlst[cc]);
  return cc ? -cc : lstsiz;
}
int populateDEVPRPIS(char *fecname,char *eqmname,char *lstname,DBLNKLIST *fecList,NAME64 **prps)
{
  xmlFIDS *fec;
  xmlXIS *eqm;
  xmlSETIS *set;
  NAME64 *member;
  int i, cc = 0, lstsiz = 0;

  if (fecList == NULL || prps == NULL) ccerr(argument_list_error);
  listGoToFirst(fecList);
  while((fec=listGetNext(fecList)) != NULL) 
  { /* find the fec */
    if (strnicmp(fec->name,fecname,FEC_NAME_SIZE) == 0)
    { /* that's the one */
      listGoToFirst( fec->eqms );
      while((eqm=listGetNext(fec->eqms)) != NULL) 
      { /* find the equipment module */
        if (strncmp(eqm->name,eqmname,EQM_NAME_SIZE)) continue;
        listGoToFirst(eqm->sets);
        while ((set=listGetNext(eqm->sets)) != NULL)
        {
          if (strncmp(set->name,lstname,SET_NAME_SIZE)) continue;
          /* that's the set !*/
          lstsiz = set->members->nElements;
          listGoToFirst(set->members);
          if ((*prps=(NAME64 *)SystemCalloc(lstsiz,sizeof(DEVIS))) == NULL) ccerr(out_of_server_memory);
          for (i=0,listGoToFirst(set->members); i<lstsiz && (member=(NAME64 *)listGetNext(set->members)) != NULL; i++)
          { /* run through eqm->devices for the common devices */
            strncpy((*prps)[i].name,member->name,PROPERTY_NAME_SIZE);
          }
        }
      }
    }
  }
err:
  if (cc) feclog("populate Export Info Database from XML : %s",erlst[cc]);
  return cc ? -cc : lstsiz;
}
int populateAIS(char *fecname,DBLNKLIST *fecList,ALIAS_TABLE **ais)
{
  xmlFIDS *fec;
  xmlAIS *alias;
  int i, cc = 0, lstsiz = 0;

  if (fecList == NULL || ais == NULL) ccerr(argument_list_error);
  listGoToFirst(fecList);
  while((fec=listGetNext(fecList)) != NULL) 
  { /* find the fec */
    if (strnicmp(fec->name,fecname,FEC_NAME_SIZE) == 0)
    { /* that's the one */
      listGoToFirst( fec->alias );
      while((alias=listGetNext(fec->alias)) != NULL) lstsiz++;
      if ((*ais=(ALIAS_TABLE *)SystemCalloc(lstsiz,sizeof(ALIAS_TABLE))) == NULL) ccerr(out_of_server_memory);
      for (i=0,listGoToFirst(fec->alias); i<lstsiz && (alias=(xmlAIS *)listGetNext(fec->alias)) != NULL; i++)
      { /* run through eqm->devices for the common devices */
        strncpy((*ais)[i].alias,alias->name,DEVICE_NAME_SIZE);
        strncpy((*ais)[i].name,alias->target,DEVICE_NAME_SIZE);
      }
    }
  }
err:
  if (cc) feclog("populate Export Info Database from XML : %s",erlst[cc]);
  return cc ? -cc : lstsiz;
}
int xmlReadFile(char *filepath, char *filename, DBLNKLIST *fecList)
{
  FILE *h;
  char c;
  char fn[256], data[256], tag[32], unknownTag[32];
  int  cc = 0;
  int  dataCharCnt = 0;
  int  tagCharCnt = 0;
  int  tagMarkerFound = 0;
  int  closingTagOpened = 0, openingTagOpened = 0;
  int  dataAllowed = 0;	 /* if the tag is opened <tag> and not closed yet </tag> */

  memset( unknownTag, '\0', 32 );
  sprintf(fn,"%s%s",filepath,filename);
  if( (h = fopen( fn, "r" )) == NULL ) ccerr(no_such_file);

  DocumentHandler.currentDataSection = fecList;
  setCurrentHandler( &DocumentHandler );

  while( !feof(h) )
  {
    c = getc(h);
    switch( c )
    {
      case '<': 
        data[dataCharCnt] = 0;
        if( dataAllowed ) CurrentHandler->processData( data );
        dataAllowed = 0;
        data[0] = 0;
        dataCharCnt = 0;
        tagCharCnt = 0;
        tagMarkerFound = 1;
        openingTagOpened = 1;
        if ((c=getc(h)) == '/') 
        {        
            openingTagOpened = 0;
            closingTagOpened = 1; 			
        }
        else 
        {
          ungetc(c,h);
        }
        break;
      case '>':
        if( openingTagOpened ) 
        {
          tag[tagCharCnt] = 0;
          if( !unknownTag[0] )
          {
            dataAllowed = CurrentHandler->processOpeningTag( tag );
            if( !dataAllowed ) strncpy( unknownTag, tag, 31 ); 
          }
          tag[0] = 0;
          tagCharCnt = 0;
        }
        else if( closingTagOpened ) 
        {
          tag[tagCharCnt] = 0;
          if( unknownTag[0] )
          {
            if( !strcmp( tag, unknownTag ) ) unknownTag[0] = 0;
          }
          else
          {
            CurrentHandler->processClosingTag( tag );
            tag[0] = 0;
            tagCharCnt = 0;
            dataAllowed = 0;
          }
        }
        openingTagOpened = closingTagOpened = 0;
        break;
      default:		 
      if( c >= ' ' )
      {
        if( openingTagOpened || closingTagOpened ) tag[tagCharCnt++] = c;
        else if( dataAllowed ) data[dataCharCnt++] = c;
      }		 
    }
  }
err:
  if (h != NULL) fclose(h);
  return cc;
}

void BytesToHex(BYTE *b,char *s,int n)
{
  int v;

  while (n--)
  {
    *s++ = (v = (*b&0xf0)>>4) < 10 ? '0' + v : 'A' + v - 10;
    *s++ = (v = *b++&0xf) < 10 ? '0' + v : 'A' + v - 10;
  }
  *s = 0;
  strrev(s);
  return;
}
/**
 * \internal
 * \brief Converts a string into a byte array.
 * 
 * Each digit is interpreted as a hexadecimal number.
 * 
 * \note This function also prepends up to 13 '0's to the input string. 
 * 
 * \param s String containing a hexadecimal number (input and output).
 * \param b Pointer to resulting byte array.
 * \param nibs Double length of b and ALSO buffer length (not string length) of s.
 * \todo This is two functions in one: a) Prepend zeros to a number, b) convert to hex
 */
void HexToBytes(char *s,BYTE *b,int nibs)
{
  char buf[20];
  int k,i;

  if ((k=(int)strlen(s)) < nibs)
  {
    strrev(s); strncat(s,"000000000000",nibs-k);
    strrev(s);
  }
  for (i=0; i<nibs/2; i++)
  {
    strncpy(buf,&s[i*2],2); buf[2] = 0;
    b[i] = (unsigned char)strtol(buf,NULL,16);
  }
}
unsigned int HexToDec(char *str)
{
  int i,d;
  unsigned int off,value;

  off=1;value=0;
  for (i=(int)strlen(str)-1;i>=0;i--)
  {
    if (str[i]>='A') d=str[i]-'A'+10;else d=str[i]-'0';
    value=value+off*d;
    off=off*16;
  }
  return value;
}
#if !defined(UNIX) && !defined(VXWORKS) && !defined(NIOS)
/* provide this posix routine for those OSes which don't have it */
TINE_EXPORT char *strtok_r(char *s, const char *delim, char **ptrptr)
{
  char *str, *c;
  if (delim == NULL || ptrptr == NULL) return NULL;
  if (s != NULL) /* first call */
  {
    /* start at the beginning of the string given */
    *ptrptr = s + strspn(s,delim); /* skip initial delimiters */
  }
  if (*ptrptr == NULL) return NULL; /* I must be at the end */
  str = *ptrptr; /* this is the current starting point */
  if ((c=strpbrk(str,delim)) != NULL) /* move along */
  {
    *ptrptr = c + strspn(c,delim);
    *c = 0;
  }
  else /* no more tokens */
  {
    *ptrptr = NULL;
  }
  return str;
}
#endif

/**
 * \internal
 * \brief Removes whitespace from the beginning and the end of a string.
 * 
 * Also removes newline characters from the end of the string.
 *
 * \note Modifies the input string in situ.
 * 
 * \return Pointer to modified string.
 */
TINE_EXPORT char *strtrm(char *s)
{
  char *start, *end;
  unsigned int len;

  if (s == NULL) return s;  /* NULL given, give it back */
  if ((len=(int)strlen(s)) == 0) return s; /* empty string given, give it back */
  /* find first non-whitespace */
  for (start=s; isspace(*start) && len > 0; start++, len--); 
  /* find last non-whitespace */
  for (end=start+len-1; isspace(*end) && len > 0; end--, len--) *end = 0;
  /* Move to the left */
  memmove(s, start, len + 1);
  /* return a pointer to the modified string as a curtesy */
  return s;
}

/**
 * \internal
 * \brief Counts the lines in a text file.
 * 
 * This function assumes that the first line is a header line,
 * which is not counted. It does not count comment lines.
 * Comment lines are lines which start with '%", '#', or ';'.
 * 
 * \return 
 * <ul> 
 * <li>number of lines</li>
 * <li>-1 if file is empty</li>
 * <li>0 if the file contains only a header line</li>
 * <li>-no_such_file if the file could not be openend</li>
 * </ul> 
 */
int GetDataBaseSize(char *filename)
{
  char str[256];
  int line = -1;
  FILE *fp;

  if (filename == NULL || strlen(filename) == 0) return -argument_list_error;
  if ((fp=fopen(filename,"r")) == NULL) return -no_such_file;
  while (fgets(str,255,fp) != NULL)   /* scan file       */
  {
    strtrm(str);
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#",str[0])) continue;      /* comment line(s) */
    line++;
  }
  fclose(fp);
  return line;
}
int csvNameServerFillPortOffset(char *po);
int csvNameServerFillHAddr(char *ip);
int csvNameServerFromHost(char *host);
/**
 * \internal
 * \brief Array of FEC data.
 * This array serves as a local cache for DEC addresses.
 * Each entry in FecTbl is identified by a FEC name and contains
 * the information which is needed to access that FEC.
 * 
 * Declared in srvcore.h, allocated in initFecTable(), deallocated in freeServerMemory().
 * \note Index 0 in this array always refers to the current name server. 
*/
FecDataStruct *FecTbl = NULL;

/**
 * \internal
 * \brief Array of Name server data.
 * 
 * This array contains information about available name servers.
 * 
 * Declared in srvcore.h, allocated in SetNameServerAddress(), deallocated in freeServerMemory().
 * 
 * \see numNameServers
 */
FecDataStruct *NameServer = NULL;

/**
 * \internal
 * \brief Number of valid entries in NameServer.
 * 
 * Declared in dbtools.c, srvcore.h.
 * 
 * \see NameServer
 */
int numNameServers = 0;
CSVCOLUMN csvNameServerColumns[] =
{
  { "FEC_NAME","",0,CF_TEXT,FEC_NAME_SIZE,(BYTE *)OFFSETIN(FecDataStruct,Name),NULL,0 },
  { "FEC_NETWORK","0",0,CF_BYTE,8,(BYTE *)OFFSETIN(FecDataStruct,IPXNetwork),NULL,0 },
  { "FEC_NODE","0",0,CF_BYTE,12,(BYTE *)OFFSETIN(FecDataStruct,IPXNode),NULL,0 },
  { "IP_ADDR","",0,CF_TEXT,16,(BYTE *)OFFSETIN(FecDataStruct,IP),csvNameServerFillHAddr,0 },
  { "PORT_OFFSET","0",0,CF_LONG,1,(BYTE *)OFFSETIN(FecDataStruct,PortOffset),NULL,0 },
  { "FEC_HOST","",0,CF_NULL,0,NULL,csvNameServerFromHost,0 },
};

CSVDB csvNameServerDb =
{
  csvNameServerColumns,
  sizeof(csvNameServerColumns)/sizeof(csvNameServerColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(FecDataStruct)
};

int csvNameServerFillHAddr(char *ip)
{
  char str1[32],str2[32];
  DWORD haddr = inet_addr(ip);
  memcpy(NameServer[csvNameServerDb.cnt].IPh_addr,&haddr,4);
  if (tineDebug)
  {
    BytesToHex(NameServer[csvNameServerDb.cnt].IPXNetwork,str1,4);
    BytesToHex(NameServer[csvNameServerDb.cnt].IPXNode,str2,6);
    dbglog("Using Name Server %s [%s %s %s]",
            NameServer[csvNameServerDb.cnt].Name,
            str1,str2,
            NameServer[csvNameServerDb.cnt].IP);
  }
  return 0;
}
int csvNameServerFromHost(char *host)
{
  struct hostent *hent;
  static struct sockaddr_in sa;
  int ptr = csvNameServerDb.cnt;
# if defined(WIN32)
  /* do this here, just in case this is called prior to SystemInit() ! */
  if (!WinsockLoaded()) return system_error;
# endif
  if (NameServer == NULL)
  {
    if ((NameServer=(FecDataStruct *)SystemCalloc(1,sizeof(FecDataStruct))) == NULL) return out_of_server_memory;
    numNameServers = 1;
    ptr = 0;
  }
  if (host != NULL && strlen(host) != 0 && (hent=gethostbyname(host)) != NULL)
  {
    memcpy(NameServer[ptr].IPh_addr,hent->h_addr,4);
    memcpy(&sa.sin_addr,hent->h_addr,sizeof(struct in_addr));
    inet_ntoa_b(sa.sin_addr,NameServer[ptr].IP);
    if (strlen(NameServer[ptr].Name) == 0)
    {
      sprintf(NameServer[ptr].Name,"ENS#%d",numNameServers);
    }
  }
  return 0;
}
/**
 * \internal
 * \brief Indicates if NameServer has already been initialized.
 * 
 * Roughly equivalent to NameServerLoaded.
 * \see SetNameServerAddress, GetNameServerAddress
 */
int gotNameServerAddress = 0;

/**
 * \internal
 * \brief Initialize Name server information
 * 
 * <ol>
 * <li>If the envronment variables TINE_ENS or CONTROLENS are set, use name
 * server information from there.</li>
 * <li>If not, read name server information from cshosts file which is expected to
 * be in a directory indicated by TINE_HOME or CONTROLDB. Therby changes/sets tineHomePath.</li>
 * </ol> 
 * 
 * Only called by GetAllocSizesFromNameServer()
 * 
 * \see GetAddressFromNameServer()
 * \see SetNameServerAddress()
 * \see tineHomePath
 * 
 * \todo tineHomePath is already initialized in initFecTable.
 */
int GetNameServerAddress(void)
{
  int i, cc;
  char *ptr;

  if (gIsEquipmentNameServer)
  {
    feclog("Server is the ENS: not using a name server!");
    feclog("TINE HOME : working directory");
    numNameServers = 0;
    return not_implemented;
  }
  if ((ptr=getenv("TINE_ENS")) != NULL || (ptr=getenv("CONTROLENS")) != NULL) 
    SetNameServerAddress(ptr);
  if (gotNameServerAddress) return 0;
  tineHomePath[0] = 0;
  if ((ptr=getenv("TINE_HOME")) != NULL || (ptr=getenv("CONTROLDB")) != NULL) 
    strcpy(tineHomePath,ptr);
# ifdef FS_DELIMITER
  if (((i=(int)strlen(tineHomePath)) > 0) &&
      (tineHomePath[i-1] != FS_DELIMITER))  /* no delimiter ! */
  {
    tineHomePath[i] = FS_DELIMITER; tineHomePath[i+1] = 0;
  }
# endif

  feclog("TINE HOME : [%s]",tineHomePath);
  numNameServers = 0;
  if ((cc=csvReadFile(tineHomePath,CSHOSTS_FILE,&csvNameServerDb,(void **)&NameServer)) == 0)
  {
    gotNameServerAddress = 1;
    numNameServers = csvNameServerDb.siz;
  }
  else
  { /* one last try from the DNS */
    csvNameServerFromHost("tineens");
    if (numNameServers == 0) 
    {
      feclog("name server address unresolved");
      cc = name_server_unknown;
    }
    else 
    {
      cc = 0;
      gotNameServerAddress = 1;
    }
  }
  for (i=0; i<numNameServers; i++)
  {
    NameServer[i].TineProtocol = DEFAULT_TRANSPORT_PROTOCOL_LEVEL;
    NameServer[i].TransportProtocol = UDP;
  }
  return cc;
}

/**
 * \internal
 * \brief Determines the index of a specified column name in a header line.
 * 
 * Some alternate spellings are tolerated.
 * 
 * \return Column index, starting with 0.
 * 
 */
int findcol(char *str,char *tag)
{
  char tagAlt0[96],tagAlt1[96],tagAlt2[96],altc=0,brkc=0;
  char *c;
  int len,i,k;

  if (str == NULL || tag == NULL) return -1;
  /* effect a string-insensitive compare */
  strncpy(tagAlt0,tag,96);
  strupr(tagAlt0);  
  if ((c=strchr(tagAlt0,' ')) != NULL) 
  {
    brkc = ' ';
    altc = '_';
  }
  else if ((c=strchr(tagAlt0,'_')) != NULL) 
  {
    brkc = '_';
    altc = ' ';
  }
  if (brkc != 0) /* prepare alternates */
  {
    for (i=0,k=0; i<(int)strlen(tagAlt0) && i<95; i++)
    {
      if (tagAlt0[i] != brkc)
      {
        tagAlt1[i] = tagAlt0[i];
        tagAlt2[k++] = tagAlt0[i];
      }
      else /* tagAlt0[i] == brkc */
      {
        tagAlt1[i] = altc;
      }
    }
    tagAlt1[i] = 0; tagAlt2[k] = 0;
  }
  for (i=0,c=str; strlen(c); i++,c++)
  {
    while (*c == '\"') { c = strchr(++c,'\"'); if (c++ == NULL) return -1; }
    while (isspace(*c) != 0) c++;
    strupr(c);
    len = (int)strcspn(c,",\n"); while (isspace(c[len-1]) != 0) len--;
    if (!strncmp(c,tagAlt0,MAX(len,(int)strlen(tagAlt0)))) return i;
    if (brkc != 0) 
    {
      if (!strncmp(c,tagAlt1,MAX(len,(int)strlen(tagAlt1)))) return i;
      if (!strncmp(c,tagAlt2,MAX(len,(int)strlen(tagAlt2)))) return i;
    }
    if ((c=strpbrk(c,",\n")) == NULL) return -1;
  }
  return -1;
}

/**
 * \internal
 * \brief Extracts a partial string from a comma-separated line.
 * 
 * This function scans a comma-separated line and extracts the column
 * which is indexed by col.  The extracted string is copied into
 * the supplied buffer (and is thread safe!)
 * 
 * \param col Column number. Indexing starts at 0.
 * \param str Comma-separated line.
 * 
 * \return 0 upon success or a TINE error code.
 * 
 */
int getColumnStr(int col,char *str,char *buf,int bufsiz)
{
  int colcnt=0, cc = 0;
  char *c;
  int sslen, FindEndOfString = FALSE;

  if (str == NULL || buf == NULL) ccerr(argument_list_error);

  /* find column position 'col' (count the non-embedded commas) : */
  /* c points to current position in string; DO NOT scan past the string length !*/
  for (c=str; strlen(c) > 0 && colcnt < col; c++)
  {
    /* scan through the string ... */
    if (FindEndOfString)
    {
      /* jump ahead to it */
      if ((c=strchr(c,'\"')) == NULL) ccerr(string_expected);
      FindEndOfString = FALSE; /* found it */
    }
    else
    {
      if (*c == ',') colcnt++;
      else if (*c == '\"') FindEndOfString = TRUE; /* quoted string */
    }
  }

  /* did we jump out because strlen(c) == 0 ? */
  if (colcnt < col) ccerr(no_such_column);

  while (isspace(*c) != 0) c++; /* trim leading spaces */
  if (*c == '\"') 
  {
    c++;   /* if quoted, jump over initial quote */
    sslen = (int)strcspn(c,"\""); /* length up to the next \" */  
  }
  else
  {
    sslen = (int)strcspn(c,",\n"); /* length up to the next , or \n */  
  }
  while (sslen > 0 && isspace(c[sslen-1]) != 0) sslen--; /* Strip trailing spaces */
  if (sslen > bufsiz)  
  {
    /* sslen longer than expected; write log message and continue anyway */
    feclog("sub string \"%s\" too long (%d vs %d)",c,sslen,bufsiz);
    sslen = bufsiz;
  }
  strncpy(buf,c,sslen); /* Copy sub-string to buffer */
  if (sslen <= bufsiz) buf[sslen] = 0; /* add \0 if there's room */

err:
  if (cc)
  {
    c = str == NULL ? "<null string>" : str;
    feclog("csv: column %d in %s -> %s",col,c,erlst[cc]);
  }
  return cc;
}

/**
 * \internal
 * \brief Extracts a partial string from a comma-separated line.
 * 
 * This function scans a comma-separated line and extracts the column
 * which is indexed by col.  The extracted string is copied into
 * a static buffer which is returned (not thread-safe!).
 * 
 * A column may me at most 127 bytes long or will be truncated.
 * 
 * \param col Column number. Indexing starts at 0.
 * \param str Comma-separated line.
 * 
 * \return Pointer to static buffer.
 * 
 */
char *colptr(int col,char *str)
{
  static char buf[128];
  buf[0] = 0;
  getColumnStr(col,str,buf,128);
  return buf;
}

/**
 * \internal
 * \brief Compares a column of a comma-separated line to a string.
 * 
 * This function scans a comma-separated line and extracts the column which is indexed by col. 
 * 
 * \param name String to compare.
 * \param str Comma-separated line.
 * \param col Column index (starting at 0)
 * \param limit Maximal length of the extracted column (length of result buffer).
 * 
 * \return like strcmp().
 * 
 */
int namcmp(char *name,char *str,int col,int limit)
{
  int n;
  char *scratch;
# if defined(VXWORKS)
  /*  VxWorks does not understand alloca() */
  char scratchBuffer[516];
  if (limit < 0 || limit > 512) return -1;
  scratch = scratchBuffer;
# else
  if (limit < 0) return -1;
  if ((scratch=(char *)alloca(limit)) == NULL) return -1; /* practically impossible */
# endif
  if (name == NULL && str != NULL) return -1; /* problem with arguments */
  if (str == NULL && name != NULL) return -1;
  if (name != NULL && col < 0) return -1; /* column not present */
  getColumnStr(col,str,scratch,limit); 
  n = MAX(limit,(int)strcspn(scratch,",\n"));
  return strncmp(name,scratch,n);
}

/**
 * \internal
 * \brief Standard Initializer used for most CSV database readers
 * 
 * Allocate internal buffer memory for a CSVDBStruct.
 * This function assumes that a CSVDBStruct is already partially initialized.
 * <ul>  
 * <li>It allocates memory for the buffer db->tgt if db->tgt_siz > 0. 
 * It uses the values of the 
 * fields db->siz, db->tgt_siz, and db->ncols to determine the buffer size.
 * </li>
 * <li> 
 * It also assumes that db->col is already initialized.
 * For each entry in db->col, it replaces the
 * pointer col->start with the previous value of col->ptr and
 * adds the pointer db->tgt as an offset to the pointer col->ptr.
 * </li>
 * 
 * <li> 
 * It sets db->cnt to 0. 
 * </li>
 * 
 * </ul>
 * 
 * \see CSVDBStruct
 * \todo Misleading name. Was this originally intended as (part of) an initializer ?
 * (This is THE standard initializer for most CSV files)
 * \todo What does the col pointer juggling mean (understood roughly)?
 * (if the csv contents are supposed to be filling a structure line by line, then the
 *  pointer has to be advanced).
 * \todo The concept behind CSVDBStruct and csv* functions is very hard to understand.
 * \bug Do not call csvStdInit twice (UNLESS csvStdExit is called in between) ... 
 */
TINE_EXPORT void csvStdInit(struct CSVDBStruct *db)
{
  int i;
  if (db == NULL || db->tgt_siz <= 0) return;
  db->cnt = 0;
  if (db->tgt != NULL) SystemFree(db->tgt);
  db->tgt = (BYTE *)SystemCalloc(db->siz,db->tgt_siz);
  for (i=0; i<db->ncols; i++) 
  {
    db->col[i].start = db->col[i].ptr; /* reset via csvStdExit() */
    /* set the ptr to the container addr plus the original offset */
    db->col[i].ptr = db->tgt + (size_t)db->col[i].start; 
  }
}
/**
 * \internal
 * \brief Standard Interator for most CSV database files
 *
 * Increments column index, increments column pointers.
 * \todo pointer juggling (see code)
 * \todo better: use real (separate) iterator instead of double meaning of CSVDBStruct
 *
 * \note the pointer advancement and size of the container have nothing to do with
 * column offsets and row indices.
 */
TINE_EXPORT int csvStdIterator(struct CSVDBStruct *db)
{
  int i;
  if (db == NULL || db->tgt_siz <= 0) return -1;  /* don't iterate */
  for (i=0; i<db->ncols; i++) if (db->col[i].fmt != CF_NULL) db->col[i].ptr += db->tgt_siz;
  db->cnt++;
  return 0;
}
/**
 * \internal
 * \brief Standard Finalizer for most CSV database files.
 * 
 * Restores column pointers to their values before memory allocation.
 * Has to be called before Memory-reallocation (csvStdInit()).
 * Does not de-allocate memory => Is not kind of a destructor
 * as the name suggests.
 * 
 * \todo misleading name (why?)
 * \todo pointer juggling (see csvStdIterator)  
 * \bug this function rather seems to be a work-around (see )
 * \see csvStdInit
 */
TINE_EXPORT void csvStdExit(struct CSVDBStruct *db)
{
  int i;
  if (db == NULL || db->tgt_siz <= 0) return;
  for (i=0; i<db->ncols; i++) db->col[i].ptr = db->col[i].start;
}
/**
 * \internal
 * \brief This function takes a string, converts it to a data type
 * specified by a format constant, and stores the result at a place
 * pointed to by a void pointer.
 */
TINE_EXPORT int csvAssignValue(void *ptr,int fmt,int siz,char *str)
{
  char *pstr = str;
  if (ptr == NULL) return 0;         /* intentionally not assigned */
  if (str == NULL) return argument_list_error;
  if (fmt != CF_TEXT && fmt != CF_NULL && strlen(str) == 0) pstr = "0";
  switch (fmt)
  {
    case CF_NULL: break;
    case CF_TEXT: if (siz == 0) *(char *)ptr = 0; else strncpy((char *)ptr,pstr,siz); break;
    case CF_BYTE: 
      if (siz > 1 && siz <= (int)strlen(str)) HexToBytes(pstr,(BYTE *)ptr,siz); 
      else *(BYTE *)ptr = (BYTE)atoi(pstr);
      break;
    case CF_SHORT: *(short *)ptr = (short)strtol(pstr,NULL,0); break;
    case CF_LONG:  *(SINT32 *)ptr = strtol(pstr,NULL,0); break;
    case CF_FLOAT: *(float *)ptr = (float)strtod(pstr,NULL); break;
    case CF_DOUBLE: *(double *)ptr = strtod(pstr,NULL); break;
    default: return illegal_format;
  }
  return 0;
}
/**
 * \internal
 * \brief Reads the contents of a CSV file and places the entries in a container.
 * 
 * The contents of parameter target may be NULL or a valid pointer.
 * If target is not NULL, target will be freed and overwritten with 
 * a pointer to the internal memory of the database structure.
 * If db has no "ini", target will be overwritten
 * with a pointer to db's internal memory.
 * If db has an ini function, it reallocates its internal memory. 
 * \param path The directory where the file is
 * \param file The file to open
 * \param db The db structure 
 * \param target pointer which is freed an then set to db->tgt
 * \see findcol
 * \note The memory to which target points is freed before re-allocation.
 * \todo clarify weird allocation/deallocation
 * \bug db-tgt is not de-allocated when target != NULL
 *
 * \note To date, the only csv file which is occasionally re-read and 
 * actually needs the memory re-allocation is users.csv. Otherwise,
 * making the same csvReadFile() call a second time will have no effect
 * in the end as long as the standard ini, itr, and exi routines are used.
 *
 * \note some CSV File use different (non-standard) routines. e.g. csvExportIterator().
 *
 */
TINE_EXPORT int csvReadFile(char *path,char *file,CSVDB *db,void **target)
{
  FILE *fp=NULL;
  int i,cc=0,done=0;
  char *c,str[256],hdr[256],scratch[80],fn[128];

  if (file == NULL) return argument_list_error;
  if (path == NULL) strncpy(fn,file,128); else sprintf(fn,"%s%s",path,file);
  /* find database size => effectively the number of rows */
  if ((db->siz=GetDataBaseSize(fn)) < 0) ccerr(-db->siz);
  /* was a pre-allocated target passed ? */
  if (target != NULL) db->tgt = (BYTE *)*target;
  /* call init routine if there */
  if (db->ini != NULL) db->ini(db);
  /* either the target was pre-allocated or the initializer allocated it */
  if (target != NULL) *target = db->tgt;
  /* read through */
  if ((fp=fopen(fn,"r")) == NULL) ccerr(no_such_file);
  while (fgets(str,255,fp) != NULL)
  {
    strtrm(str);
    if (strlen(str) == 0) continue;               /* carraige return */
    if (strchr("%;#\n",str[0]) != NULL) continue; /* comment line */
    if (!done)
    {
      strncpy(hdr,str,255);
      for (i=0; i<db->ncols; i++) db->col[i].pos = findcol(hdr,db->col[i].nam);
      done = TRUE;
      continue;
    }
    for (i=0; i<db->ncols; i++)
    {
      if (db->col[i].pos < 0) 
      {
        c = db->col[i].def;
      }
      else
      {
        getColumnStr(db->col[i].pos,str,scratch,64); scratch[64] = 0;
        c = scratch;
      }
      /* assign value to structure location */
      cc = csvAssignValue(db->col[i].ptr,db->col[i].fmt,db->col[i].siz,c);
      if (cc)
      {
        sprintf(str,"column %s entry %d : %s\n",db->col[i].nam,i,erlst[cc]);
        feclog(str); ttyoutput(str);
        ccerr(cc);
      }
      /* call action function if there */
      if (db->col[i].fcn != NULL) if (db->col[i].fcn(c) != 0) break;    
    }
    /* call iterator if there */
    if (db->itr != NULL) if (db->itr(db) != 0) break;
  }
  /* call end routine if there */
  if (db->exi != NULL) db->exi(db);
err:
  if (fp) fclose(fp);
  if (cc) db->siz = 0;
  return cc;
}
/**
 * \internal
 * \brief Reads data from memory and converts it to a string.
 * \param ptr Pointer in memory
 * \param fmt TINE CF data type expected at pointer
 * \param siz Maximal string size
 * \param str Destination
 */
TINE_EXPORT int csvGetValue(void *ptr,int fmt,int siz,char *str)
{
  if (ptr == NULL) return 0;         /* intentionally not assigned */
  if (str == NULL) return argument_list_error;
  switch (fmt)
  {
    case CF_TEXT: strncpy(str,(char *)ptr,siz); break;
    case CF_BYTE: BytesToHex((BYTE *)ptr,str,siz); break;
    case CF_SHORT: sprintf(str,"%d",*(short *)ptr); break;
#   ifdef OS_16BIT
    case CF_LONG:  sprintf(str,"%ld",*(SINT32 *)ptr); break;
#   else
    case CF_LONG:  sprintf(str,"%d",*(SINT32 *)ptr); break;
#   endif
    case CF_FLOAT: sprintf(str,"%g",*(float *)ptr); break;
    case CF_DOUBLE: sprintf(str,"%g",*(double *)ptr); break;
    default: return illegal_format;
  }
  return 0;
}
/**
 * \internal
 * \brief Writes data in a CSVDB to a file.
 * \param path Directory
 * \param file File name
 * \param db A CSVDB
 */
TINE_EXPORT int csvWriteFile(char *path,char *file,CSVDB *db)
{
  FILE *fp=NULL;
  int i,j,cc=0;
  char str[512],scratch[128],fn[128];

  if (file == NULL) return argument_list_error;
  if (path == NULL) strncpy(fn,file,128); else sprintf(fn,"%s%s",path,file);
  if ((fp=fopen(fn,"w")) == NULL) ccerr(file_error);
  for (i=0,str[0]=0; i<db->ncols; i++)
  {
    strcat(str,db->col[i].nam);
    if (i<db->ncols-1) strcat(str,","); 
    if (db->tgt_siz > 0)  /* return pointers to start position */
    {
      db->col[i].ptr = db->tgt + (size_t)db->col[i].start;
    }
  }
  strcat(str,"\n");
  fwrite(str,strlen(str),1,fp);
  for (j=0; j<db->siz; j++)
  {
    str[0] = 0;
    for (i=0; i<db->ncols; i++)
    {
      csvGetValue(db->col[i].ptr,db->col[i].fmt,db->col[i].siz,scratch);
      strncat(str,scratch,db->col[i].siz);
      if (i<db->ncols-1) strcat(str,","); 
    }
    strcat(str,"\n");
    fwrite(str,strlen(str),1,fp);
    /* call iterator if there */
    if (db->itr != NULL) if (db->itr(db) != 0) break;
  }
  /* call end routine if there */
  if (db->exi != NULL) db->exi(db);
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int makeDirectoryPath(char *dname)
{
# ifndef FS_RDONLY
# if defined(UNIX) || defined(WIN32)
# if defined(UNIX)
  DIR *dp;
  char *tok = "/";
# elif defined(WIN32)
  char fn[128];
  char *tok = "\\";
  struct _finddata_t fp;
# endif
  char *c=dname,*cp=NULL,tmpname[128],dirname[128];

  strncpy(tmpname,dname,128); dirname[0] = 0;
  while (*c++ == *tok) strcat(dirname,tok); /* in case it begins with '/' or '\\' */
  for (c=strtok_r(tmpname,tok,&cp); c != NULL; c=strtok_r(NULL,tok,&cp))
  {
    if (strlen(c) == 0) break; /* terminating delimiter gives the last token empty */
    strcat(dirname,c);
    strcat(dirname,tok);
#   if defined(UNIX)
    if ((dp=opendir(dirname)) != NULL)
    {
      closedir(dp);
    }
    else
    {
      if (mkdir(dirname,S_IWUSR|S_IRUSR|S_IXUSR|S_IWGRP|S_IRGRP|S_IXGRP|S_IROTH)) return 0;
    }
#   elif defined(WIN32)
    sprintf(fn,"%s*.*",dirname);
    if (_findfirst(fn, &fp) == -1)
    {
      if (mkdir(dirname)) return 0;
    }
#   endif
  }
# endif
# else /* FS_RDONLY */
  dname = dname;
# endif
  return -1;
}
#endif /* NO_DOXYGEN */
