/** \file database.c
 * \brief Database routines used for name resolution, etc.
 *      
 */
#include "prolog.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "dbtools.h"
#include "srvcore.h"
#include "srvip.h"
#include "tinetype.h"
#include "errors.h"
#include "globals.h"


int gIsEquipmentNameServer = FALSE;
int gUsingLegacyAddrCalls = FALSE;

int prepIncomingAddress(int isLegacy,char *addrstream,FecDataStruct *fec,ExpDataStruct *srv);
int queryNameServerForFec(char *fecName,FecDataStruct *fec);

TINE_EXPORT void isEquipmentNameServer(int value) { gIsEquipmentNameServer = value; }
static BYTE zero[6] = {0,0,0,0,0,0};
/** 
 * \internal
 * \brief General client-side database path
 * 
 *  Given by the environment variable setting "TINE_HOME" or (deprecated) 
 *  CONTROLDB.  This specifies the location of the 'cshost.csv' file (among others).
 *
 *  Declared in dbtools.h, initialized in initFecTable().
 *
 * \todo refacter the name to TineHomePath
 *
 */
char tineHomePath[80];

/**
 * \internal
 * \brief Number of valid entries in SrvTbl
 * \see SrvTbl
 */
int numSrvTblEntries=0;

/**
 * \internal
 * \brief Number of valid entries in FecTbl
 * \see FecTbl
 */
int numFecTblEntries=0;

/**
 * \internal
 * \brief Maximum size of the SrvTbl table
 *
 * This is more or less hard-wired to the MaxNumConnections.  The
 * original intent was to let this grow dynamically as needed, but
 * there are some unfortunate side-effects on some OSes without good
 * MMUs.  So the value is currently being set up front.
 *
 * \see SrvTbl
 */
int SrvTblSize = 0;

/**
 * \internal
 * \brief Maximum size of the FecTbl table
 *
 * This is more or less hard-wired to the MaxNumConnections.  The
 * original intent was to let this grow dynamically as needed, but
 * there are some unfortunate side-effects on some OSes without good
 * MMUs.  So the value is currently being set up front.
 *
 * \see FecTbl
 */
int FecTblSize=0;

/**
 * \internal
 * \brief Flag whether at least one ENS is known or not.
 */
int NameServerLoaded = 0;

/**
 * \internal
 * \brief Index of the currently active ENS 
 *
 * Array index of NameServer[] array which is currently
 * active.
 */
int curNameServer = 0;

/** 
 * \internal
 * \brief Array of Device Server Structs.
 * 
 * This array is used as a local cache which stores
 * information of known EqmContext/ExportName pairs. 
 * The array is used to map EqmContext/ExportName to
 * Equipment name, FEC name and other data.  
 * 
 * Declared in srvcore.h. 
 * Allocated in initFecTable().
 * 
 * \note The element with index 0 always refers to the current name server.
 */
ExpDataStruct *SrvTbl = NULL;

/** 
 * \internal
 * \brief ExpDataStruct which holds the ENS data
 *
 * This is the static information which is used as the 0th
 * entry of the SrvTbl table when at least one ENS is configured.
 *
 * \note If there is an ENS, then it is the 0th element in the cached
 * address tables AND the 0th element in the connection table.
 *
 */
static ExpDataStruct NameServerData; /**< Name Server data.*/

/** 
 * \internal
 * \brief Connection Table Entry which holds the ENS data
 *
 * This is the static information which is used as the 0th
 * entry of the connection table when at least one ENS is configured.
 *
 * \note If there is an ENS, then it is the 0th element in the cached
 * address tables AND the 0th element in the connection table.
 *
 */
ConnTableEntry connNameServer;

/** 
 * \internal
 * \brief Flag which determines whether to use the loopback address when client and
 * server are on the same computer.
 *
 * Some OSes (UNIX, Windows) will use a UNIX socket (pipe) or MMF as a transport
 * between client and server if it is determined that the client is on the same 
 * machine as the server.  This can be over-ridden if gUseLoopback is set to TRUE.
 *
 */
int gUseLoopback = FALSE;
TINE_EXPORT void SetUseLoopback(int value)
{
  gUseLoopback = value;
}
TINE_EXPORT int GetUseLoopback(void)
{
  return gUseLoopback;
}

/** 
 * \internal
 * \brief Determine how best to communicate with this FEC
 *
 * Sets the .TransportProtocol field to the best available communication protocol.
 *
 * \note If .TransportProtocol = 0, the address will again be queried from the ENS.
 * Thus a series of timeouts as seen in the TINE engine will set the .TransportProtocol
 * field explicitly to 0 and thereby force an address (re)-resolution.
 *
 */
int fixFecProtocol(FecDataStruct *fec)
{
  int defaultProtocol = 0;
# if !defined(VXWORKS)
  int myip = getmyipaddr();
# else
  int myip = 0;
# endif
# if defined(MAKE_TINE32_DLL)
  int hWnd;
# endif
# if defined(IPXNW) || defined(IPXBSD)
  int TransportTime = 0;
  if (memcmp(fec->IPXNetwork,zero,4))
  {
    defaultProtocol = IPX;
#   ifdef IPXNW
    if (IPXGetLocalTarget(fec->IPXNetwork,
        &fec->IPXImmediateAddress,
        &TransportTime)) return ipx_socket_error;
#   endif
  }
# endif
# if !defined(VXWORKS)
  if (myip == 0) feclog("cannot determine local ip address!");
# else
  myip = myip;
# endif
  if (memcmp(fec->IPh_addr,zero,4) || /* most often the case : a known IP address */
      !strcmp(fec->Name,SRVFEC_NAME)) /* a multicast for a service fec -> must be UDP */
    defaultProtocol = UDP;
# if defined(UNIX) && !defined(NO_TINE_PIPE)
  if (!gUseLoopback && (!memcmp(&myip,fec->IPh_addr,4) || !strncmp(fec->IP,LOOPBACKADDR,ADDR_SIZE)))
  {
    memset(fec->IPXNetwork,0,4);
    defaultProtocol = PIPE;
  }
# endif
# if defined(MAKE_TINE32_DLL) 
  if (!gUseLoopback && (!memcmp(&myip,fec->IPh_addr,4) || !strncmp(fec->IP,LOOPBACKADDR,ADDR_SIZE)))
  {
    if ((hWnd=(int)FindWindow(fec->Name,fec->Name)) != (int)NULL)
    {
      memcpy(fec->IPXImmediateAddress,&hWnd,4); 
      defaultProtocol = MMF;
    }
  }
# endif
  if (fec->TineProtocol == 0 || fec->TineProtocol > DEFAULT_TRANSPORT_PROTOCOL_LEVEL) 
    fec->TineProtocol = DEFAULT_TRANSPORT_PROTOCOL_LEVEL;
  fec->TransportProtocol = defaultProtocol;
  if (defaultProtocol)
  {
    if (tineDebug)
      dbglog("Communicating with %s via %s\n>",fec->Name,getProtocolAsString(defaultProtocol));
    return 0;
  }
  return illegal_protocol;
}

/**
 * \internal
 * \brief Locate an export name and context in the FEC table.
 * 
 * This function tries to locate an entry in the address cache: 
 * FecTbl[] and SrvTbl[] according to the specified export name and context.
 * Fetches FEC data by calling GetAddressFromNameServer() and/or GetAddressFromNameFile(), which 
 * creates or overwrites an entry in FecTbl and/or SrvTbl and return its index.
 * 
 * \param expName Exported device server name
 * \param context EqmContext
 * \param srvidx [output] index of corresponding entry in SrvTbl[] or -1. May be set to NULL if no result desired.
 * \return index in FecTbl[] or negative TINE return code.
 * \todo make expName and context const 
 * \see GetAddressFromNameServer 
 * \see GetAddressFromNameFile
 */
int locateExport(char *expName,char *context,int *srvidx)
{
  char fecName[FEC_NAME_SIZE];
  int i,cc=address_unknown,idx;
  memset(fecName,0,FEC_NAME_SIZE); /* leave fecName in the argument lists below for now */
  if ((i=GetAddressFromNameServer(context,expName,fecName,&idx)) < 0)
  {
    if (i == -link_not_open || i == connection_timeout) cc = -i;
#   ifndef FS_RDONLY
    if ((i=GetAddressFromNameFile(gAddrCacheFilePath,context,expName,fecName,&idx)) < 0)
#   endif
    if ((i=GetAddressFromNameFile(tineHomePath,context,expName,fecName,&idx)) < 0) return -cc;
  }
  if ((cc=fixFecProtocol(&FecTbl[i])) != 0) return -cc;
  if (srvidx != NULL) *srvidx = idx;
  return i;
}

/**
 * \internal
 * \brief Retrieves the fec index for the specified FEC name in the FEC table
 *
 */
int getFecIndex(char *fecName)
{
  int i;
  if (strcmp(fecName,"ENS") == 0) return 0;
  if (strcmp(fecName,"GENS") == 0) return 1;
  for (i=0; i<numFecTblEntries && strnicmp(fecName,FecTbl[i].Name,FEC_NAME_SIZE); i++);
  if (i < numFecTblEntries) return i;
  return -non_existent;
}
TINE_EXPORT int GetHostFromFecName(char *fecName,char *hostname) 
{
  int i, len = 4,cc=0;
  struct hostent *hst;
  FecDataStruct fec, *pFec;

  if ((i=getFecIndex(fecName)) < 0) 
  {
    if ((cc = -i) == non_existent)
    { /* not yet cached ? */
      cc = queryNameServerForFec(fecName,&fec);
    }
    if (cc != 0) ccerr(cc);
    pFec = &fec;    
  }
  else
  {
    pFec = &FecTbl[i];
  }
  if ((hst=gethostbyaddr((char *)pFec->IPh_addr,len,AF_INET)) == NULL) ccerr(host_not_resolved);
  strcpy(hostname,hst->h_name);
err:
  return cc;
}

/**
 * \internal
 * \brief Switches to the next name server.
 * 
 * Increments the global variable curNameServer and
 * copies data from NameServer[curNameServer] to FecTbl[0] and SrvTbl[0].
 * \see curNameServer
 * \see NameServer
 * \see FecTbl
 * \see SrvTbl
 */
TINE_EXPORT void ToggleNameServer(void)
{
  curNameServer = (curNameServer+1) % numNameServers;
  memcpy(&FecTbl[0],&NameServer[curNameServer],sizeof(FecDataStruct));
  strncpy(SrvTbl[0].FecName,NameServer[curNameServer].Name,FEC_NAME_SIZE);
  memcpy(&FecTbl[1],&NameServer[curNameServer],sizeof(FecDataStruct));
  strcpy(FecTbl[1].Name,"GENS");
  FecTbl[1].PortOffset += 101;
  feclog("ENS: toggle name server");
  msglog(0,"ENS: toggle name server");
}

int fecNameId = -1;
char fecNameTarget[FEC_NAME_SIZE+4];
char exportNameTarget[EXPORT_NAME_SIZE+4];
BYTE fecAddressBuf[sizeof(ExpDataStruct)+sizeof(FecDataStruct)];
void fillinExportName(int id,int cc)
{
  int i,j,k;
  static int ens_to = 0;
  char preCallExp[EXPORT_NAME_SIZE+4],preCallEqm[EQM_NAME_SIZE];
  FecDataStruct fec;
  ExpDataStruct srv;
  fecNameId = -1;             /* release semaphore */
  if (tineDebug > 1) dbglog("asyncLocateTag callback %d <%d>",id,cc);
  for (j=0; j<numSrvTblEntries && strnicmp(exportNameTarget,SrvTbl[j].ExportName,EXPORT_NAME_SIZE); j++);
  if (j == numSrvTblEntries)
  {
    if (numSrvTblEntries < SrvTblSize) numSrvTblEntries++;
    else return;
  }
  if (cc == 0)
  {
    prepIncomingAddress(gUsingLegacyAddrCalls,(char *)fecAddressBuf,&fec,&srv);
    ens_to = 0;
    strncpy(preCallExp,SrvTbl[j].ExportName,EXPORT_NAME_SIZE); preCallExp[EXPORT_NAME_SIZE] = 0;
    strncpy(preCallEqm,SrvTbl[j].EqmName,EQM_NAME_SHORTSIZE); preCallEqm[EQM_NAME_SHORTSIZE] = 0;
    memcpy(&SrvTbl[j],&srv,sizeof(ExpDataStruct));
    if (strnicmp(preCallExp,SrvTbl[j].ExportName,EXPORT_NAME_SIZE) || strncmp(preCallEqm,SrvTbl[j].EqmName,EQM_NAME_SHORTSIZE))
    {
      feclog("ENS: name change in assignment %s %s vs %s %s\n",
        preCallExp,preCallEqm,SrvTbl[j].ExportName,SrvTbl[j].EqmName);
    }
    for (i=0; i<numFecTblEntries && strnicmp(SrvTbl[j].FecName,FecTbl[i].Name,FEC_NAME_SIZE); i++);
    if (i == numFecTblEntries)
    {
      if (numFecTblEntries < FecTblSize) numFecTblEntries++;
      else return;
    }
    memcpy(&FecTbl[i],&fec,sizeof(FecDataStruct));
    fixFecProtocol(&FecTbl[i]);
    for (k=0; k<nConnectionTableEntries; k++)
    {
      if (conTbl[k]->srvIdx == j && !strnicmp(conTbl[k]->expName,SrvTbl[j].ExportName,EXPORT_NAME_SIZE))
      {
        conTbl[k]->fecIdx = i;  /* The device server got assigned to a new FEC ! */
      }
    }
    feclog("ENS: %s [FEC %s] added",exportNameTarget,FecTbl[i].Name);
  }
  else if (cc == illegal_read_write)
  {
    gUsingLegacyAddrCalls = TRUE;
  }
  else
  {
    if (ens_to++ > 5) ToggleNameServer();
  }
  return;
}
int asyncLocateExport(char *expName,char *context)
{
  int cc,siz;
  DTYPE dout;
  char scratch[64],cntxt[CONTEXT_NAME_SIZE+4] = "";

  if (!NameServerLoaded || fecNameId != -1) return -not_ready;

  if (tineDebug > 1) dbglog("Locate export %s",expName);
  strncpy(exportNameTarget,expName,EXPORT_NAME_SIZE);

  if (context != NULL) strncpy(cntxt,context,CONTEXT_NAME_SIZE);
  if (strlen(cntxt) == 0) sprintf(cntxt,"#2");
  sprintf(scratch,"%s/%s",CSHOSTS,cntxt);

  siz = gUsingLegacyAddrCalls ? RPCFECSTRUCT_SIZE+RPCDATASTRUCT_SIZE : FECDATASTRUCT_SIZE+EXPDATASTRUCT_SIZE;
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = (UINT32)siz;
  memset(dout.dTag,0,TAG_NAME_SIZE);
  dout.data.vptr = fecAddressBuf;

  if ((fecNameId=AttachLink(scratch,expName,&dout,NULL,CA_READ,500,fillinExportName,CM_SINGLE)) < 0)
  {
    cc = -fecNameId;
    fecNameId = -1;
    return cc;
  }
  return fecNameId;
}
#ifndef DOS
int appendAddressToCache(char *fecName,char *ip, short portOffset)
{
  int i,cc = 0;
  DWORD haddr;

  initClient();
  for (i=0; i<numFecTblEntries && strnicmp(fecName,FecTbl[i].Name,FEC_NAME_SIZE); i++);
  if (i < numFecTblEntries && FecTbl[i].TransportProtocol) ccerr(0);
  strncpy(FecTbl[i].Name,fecName,FEC_NAME_SIZE);
  strncpy(FecTbl[i].IP,ip,16);
  haddr = inet_addr(FecTbl[i].IP);
  memcpy(FecTbl[i].IPh_addr,&haddr,4);
  FecTbl[i].PortOffset = portOffset;
  fixFecProtocol(&FecTbl[i]);
  numFecTblEntries++;
err:
  return cc;
}
int appendServerToCache(char *fecName,char *context,char *expName, char *eqmName)
{
  int i,cc = 0;

  if (fecName == NULL || strlen(fecName) == 0) ccerr(argument_list_error);
  if (expName == NULL || strlen(expName) == 0) ccerr(argument_list_error);
  if (eqmName == NULL || strlen(eqmName) == 0) ccerr(argument_list_error);
  initClient();
  for (i=0; i<numSrvTblEntries; i++)
  {
    if (context != NULL && strlen(context) &&
        strnicmp(context,SrvTbl[i].EqmContext,CONTEXT_NAME_SIZE)) continue;
    if (!strnicmp(expName,SrvTbl[i].ExportName,EXPORT_NAME_SIZE)) break;
  }
  if (i < numSrvTblEntries) ccerr(0);
  strncpy(SrvTbl[i].FecName,fecName,FEC_NAME_SIZE);
  if (context != NULL) strncpy(SrvTbl[i].EqmContext,context,CONTEXT_NAME_SIZE);
  strncpy(SrvTbl[i].ExportName,expName,EXPORT_NAME_SIZE);
  strncpy(SrvTbl[i].EqmName,eqmName,EQM_NAME_SHORTSIZE);
  numSrvTblEntries++;
err:
  return cc;
}
TINE_EXPORT void FlushFecAddrCache(void)
{
	numFecTblEntries = 2; /* always leave room for ENS and GENS */
	numSrvTblEntries = 2;
	memset(&FecTbl[numFecTblEntries],0,(FecTblSize-numFecTblEntries)*sizeof(FecDataStruct));
	memset(&SrvTbl[numSrvTblEntries],0,(SrvTblSize-numSrvTblEntries)*sizeof(ExpDataStruct));
}
TINE_EXPORT void flushFecAddrCache(void) { FlushFecAddrCache(); }
TINE_EXPORT int GetENSAddr(char *name,char *addr)
{
  if (NameServerLoaded)
  {
    strncpy(name,NameServer[curNameServer].Name,FEC_NAME_SIZE);
    strncpy(addr,NameServer[curNameServer].IP,ADDR_SIZE);
  }
  else
  {
    strncpy(name,"LOCAL FILE",16);
    strncpy(addr,tineHomePath,48);
  }
  return numNameServers;
}
TINE_EXPORT int GetFECAddr(char *fecName,FecDataStruct *fecaddr)
{
  int i;
  if (fecaddr == NULL || fecName == NULL) return argument_list_error;
  for (i=0; i<numFecTblEntries && strnicmp(fecName,FecTbl[i].Name,FEC_NAME_SIZE); i++);
  if (i == numFecTblEntries) return host_not_resolved;
  memcpy(fecaddr,&FecTbl[i],sizeof(FecDataStruct));
  return 0;
}
TINE_EXPORT int GetServerAddr(char *context,char *expName,ExpDataStruct *srvaddr,FecDataStruct *fecaddr)
{
  int sidx,fidx;
  if (expName == NULL || fecaddr == NULL || srvaddr == NULL) return argument_list_error;

  for (sidx=0; sidx<numSrvTblEntries; sidx++)
  {
    if (context != NULL && strlen(context) &&
        strnicmp(context,SrvTbl[sidx].EqmContext,CONTEXT_NAME_SIZE)) continue;
    if (!strnicmp(expName,SrvTbl[sidx].ExportName,EXPORT_NAME_SIZE)) break;
  }
  if (sidx == numSrvTblEntries) return non_existent_elem;
  for (fidx=0; fidx<numFecTblEntries && strnicmp(SrvTbl[sidx].FecName,FecTbl[fidx].Name,FEC_NAME_SIZE); fidx++);
  if (fidx == numFecTblEntries) return host_not_resolved;
  memcpy(srvaddr,&SrvTbl[sidx],sizeof(ExpDataStruct));
  memcpy(fecaddr,&FecTbl[fidx],sizeof(FecDataStruct));
  return 0;
}
TINE_EXPORT int GetConnectionAddr(char *context,char *expName,char *devName,char *prpName)
{
  int cc = redirectLink(context,expName,prpName,devName);
  return cc ? CE_REDIRECTED : 0;
}
#endif

int prepIncomingAddress(int isLegacy,char *addrstream,FecDataStruct *fec,ExpDataStruct *srv)
{
  RPCFecStruct lfec;
  RPCDataStruct lsrv;
  if (fec == NULL || srv == NULL) return argument_list_error;
  if (isLegacy)
  {
#   ifdef INTEL
    memcpy(&lsrv,addrstream,RPCDATASTRUCT_SIZE);
    /* avoid alignment problems on some machines: */
    memcpy(addrstream,&addrstream[RPCDATASTRUCT_SIZE],RPCFECSTRUCT_SIZE);
    memcpy(&lfec,addrstream,RPCFECSTRUCT_SIZE);
#   else
    _StructSwap(RPCDataStructStruct,(unsigned char *)addrstream,(unsigned char *)&lsrv,RTOL);
    /* avoid alignment problems on some machines: */
    memcpy(addrstream,&addrstream[RPCDATASTRUCT_SIZE],RPCFECSTRUCT_SIZE);
	  _StructSwap(RPCFecStructStruct,(unsigned char *)addrstream,(unsigned char *)&lfec,RTOL);
#   endif
    memset(fec,0,sizeof(FecDataStruct));
    strncpy(fec->Name,lfec.Name,FEC_NAME_SIZE);
    memcpy(fec->IPXNetwork,lfec.Network,16);
    strncpy(fec->IP,lfec.IP,ADDR_SIZE);
    memcpy(fec->IPh_addr,lfec.IPh_addr,4);
    fec->PortOffset = (int)((lfec.Port[1] << 8) | lfec.Port[0]);
    if (fec->PortOffset == 0) fec->PortOffset = (int)lfec.PortOffset;
    fec->TransportProtocol = lfec.Prot;
    fec->TineProtocol = 0;
    memset(srv,0,sizeof(ExpDataStruct));
    strncpy(srv->FecName,lsrv.FecName,FEC_NAME_SIZE);
    strncpy(srv->SubSystem,lsrv.SubSystem,SUBSYSTEM_NAME_SHORTSIZE);
    strncpy(srv->ExportName,lsrv.ExportName,EXPORT_NAME_SHORTSIZE);
    strncpy(srv->EqmName,lsrv.EqmName,EQM_NAME_SHORTSIZE+1);
    strncpy(srv->EqmContext,lsrv.EqmContext,CONTEXT_NAME_SIZE);
  }
  else
  {
    memcpy(srv,addrstream,EXPDATASTRUCT_SIZE);
    /* avoid alignment problems on some machines: */
    memcpy(addrstream,&addrstream[EXPDATASTRUCT_SIZE],FECDATASTRUCT_SIZE);
#   ifdef INTEL
    memcpy(fec,addrstream,FECDATASTRUCT_SIZE);
#   else
	  _StructSwap(FDSStructStruct,(unsigned char *)addrstream,(unsigned char *)fec,RTOL);
#   endif
  }
  return 0;
}
int queryNameServerForFec(char *fecName,FecDataStruct *fec)
{
  ExpDataStruct srv;
  DTYPE dout;
  int cc=0;

  if (fecName == NULL || fec == NULL) return argument_list_error; 
  isNameServerRequest = TRUE;
retryQuery:
  if (!gUsingLegacyAddrCalls)
  { /* modern call */
    dout.dArrayLength = FECDATASTRUCT_SIZE;
    dout.dFormat = CF_STRUCT;
    memset(dout.dTag,0,TAG_NAME_SIZE);
    dout.data.bptr = &fecAddressBuf[EXPDATASTRUCT_SIZE];
    cc = tExecLink("ENS",CSHOSTS,"FEC",fecName,&dout,NULL,CA_READ,500);
    if (cc == illegal_read_write)
    {
      gUsingLegacyAddrCalls = TRUE;
      goto retryQuery;
    }
  }
  else
  { /* legacy ENS call: */
    dout.dArrayLength = RPCFECSTRUCT_SIZE;
    dout.dFormat = CF_STRUCT;
    memset(dout.dTag,0,TAG_NAME_SIZE);
    dout.data.bptr = &fecAddressBuf[RPCDATASTRUCT_SIZE];
    isNameServerRequest = TRUE;
    cc = tExecLink("ENS",CSHOSTS,"FEC",fecName,&dout,NULL,CA_READ,500);
  }
  isNameServerRequest = FALSE;      
  if (cc == 0)
  { /* throw-away srv info */
    memset(&srv,0,sizeof(ExpDataStruct));
    prepIncomingAddress(gUsingLegacyAddrCalls,(char *)fecAddressBuf,fec,&srv);
  }
  return cc;
}
int queryNameServerForAddress(char *ctx,char *exp,FecDataStruct *fec,ExpDataStruct *srv)
{
  DTYPE dout;
  int cc=0;

  if (exp == NULL || fec == NULL || srv == NULL) return argument_list_error; 
  if (!stricmp(exp,"ENS") && (strlen(ctx) == 0 || !stricmp(ctx,"SERVICE")))
  { /* an explicit call to the ENS got thru */
    memcpy(fec,&FecTbl[0],sizeof(FecDataStruct));
    memcpy(srv,&SrvTbl[0],sizeof(ExpDataStruct));
  }
  isNameServerRequest = TRUE;
retryQuery:
  if (!gUsingLegacyAddrCalls)
  { /* modern call */
    dout.dArrayLength = FECDATASTRUCT_SIZE+EXPDATASTRUCT_SIZE;
    dout.dFormat = CF_STRUCT;
    memset(dout.dTag,0,TAG_NAME_SIZE);
    dout.data.bptr = fecAddressBuf;
    cc = tExecLink("ENS",CSHOSTS,ctx,exp,&dout,NULL,CA_READ,500);
    if (cc == illegal_read_write)
    {
      gUsingLegacyAddrCalls = TRUE;
      goto retryQuery;
    }
  }
  else
  { /* legacy ENS call: */
    dout.dArrayLength = RPCFECSTRUCT_SIZE+RPCDATASTRUCT_SIZE;
    dout.dFormat = CF_STRUCT;
    memset(dout.dTag,0,TAG_NAME_SIZE);
    dout.data.bptr = fecAddressBuf;
    isNameServerRequest = TRUE;
    cc = tExecLink("ENS",CSHOSTS,ctx,exp,&dout,NULL,CA_READ,500);
  }
  isNameServerRequest = FALSE;      
  if (cc == 0)
  {
    prepIncomingAddress(gUsingLegacyAddrCalls,(char *)fecAddressBuf,fec,srv);
  }
  return cc;
}

TINE_EXPORT FecDataStruct *GetFecDataFromCache(char *context,char *expName)
{
  int fecidx,srvidx;
  static FecDataStruct fec;
  fecidx = locateExport(expName,context,&srvidx);
  if (fecidx < 0) return NULL;
  /* never return the original to the caller ! */
  memcpy(&fec,&FecTbl[fecidx],sizeof(FecDataStruct));
  return &fec;
}
/**
 * \internal
 * \brief Retrieve FEC data from name server or the local cache.
 *
 * This function first attempts to find the requested server name in its local
 * address cache.  If an complete entry is found (including transport protocol)
 * the entry's index is returned.
 * If not, the function requests the FEC data from the name server. 
 * If multiple name servers are known, it takes the first one which returns a result.
 * If the FEC is already buffered locally in FecTbl, the buffer is overweritten. 
 * Otherwise a new entry is created. If the FEC is already buffered locally in SrvTbl, 
 * the buffer is overwritten. Otherwise a ne entry is created.  
 * 
 * Special case: If the supplied export name equals "NETWORK", this indicates a (multicast)
 * request for the targetted server to respond directly.  This is for cases where the
 * ENS address is not known (diskless server, with no apriori knowledge of the ENS).
 * Usually, this technique is used to find the ENS itself.  Once known, address resolution
 * can proceed normally.  Note that server must be configured to respond to NETWORK
 * address requests, and that the ENS is usually the only server prepared to do so.
 * Also note that there are in such cases potentially two new entries in FecTbl as well as in SrvTbl. 
 * If the initially request is indeed to find the ENS, then the special slot 0 is filled with the
 * returned address.  Irregardless of this, the next available slot is filled with the full address
 * information of the responding server.
 * 
 * \param context EqmContext 
 * \param expName Export name
 * \param fecName Returns FEC name or empty string. 
 * \param srvidx  Not used if set to NULL. Otherwise, either returns the index of the corresponding 
 * element in SrvTbl[] or -1. 
 * \return I. <0 means error code. II. 0 has duplicate meaning:
 * a) fecName was found in buffer, but FEC has Protocol 0, b) FEC Index 0.
 * III. >=0 means FEC index in FecTbl[];
 * 
 * \see SrvTbl
 * \see FecTbl
 * 
 * \todo Remove gotos and jumps. (unless they're needed)
 * \todo context and expName should be const (not changed by this function).
 * \todo FecTbl has limited memory & is never cleaned => may reach limit if many FECs are used.
 * \todo Same for SrvTbl.
 */
int GetAddressFromNameServer(char *context,char *expName,char *fecName,int *srvidx)
{
  int retry = 1;  /* Jump-to-err-then-jump-back-counter; max.number of backjumps for the whole loop */
  int n,i=0,j,cc=0,isNew=FALSE,isServiceRequest=FALSE;
  SINT32 ip;
  char cntxt[CONTEXT_NAME_SIZE+4] = "";
  FecDataStruct fec;
  ExpDataStruct srv;

  if (srvidx != NULL) *srvidx = -1;
  if (!strcmp(expName,SRVEXP_NAME)) isServiceRequest = TRUE;
  if (!NameServerLoaded && !isServiceRequest) return -not_ready;
  if (expName == NULL || strlen(expName) == 0) return -argument_list_error;

  if (!stricmp(expName,"ENS") && (context == NULL || strlen(context) == 0 || !stricmp(context,"SERVICE")))
  { /* explicit call to ENS */
    if (srvidx != NULL) *srvidx = 0;
    return 0;
  }
  
  for (n=0; n<numNameServers || isServiceRequest; n++)
  { /* loop over known name servers until address is resolved */
retryRequest:
    fecName[0] = 0; /* overwrite fecName with null string */
    /* if expName == SRVEXP_NAME this is a signal to find the address via multicast */
    if (tineDebug && isServiceRequest) dbglog("using network services as FEC address");    
    /* Search SrvTbl for matching context + export name  */
    for (j=0; j<numSrvTblEntries; j++)
    { /* if context is given, then it has to match; otherwise ignore it */
      if (context != NULL && strlen(context) &&
          strnicmp(context,SrvTbl[j].EqmContext,CONTEXT_NAME_SIZE)) continue;
      if (!strnicmp(expName,SrvTbl[j].ExportName,EXPORT_NAME_SIZE)) break;
    }
    if (j < numSrvTblEntries)
    { /* EQM Entry found: Search FecTbl for matching FEC name */
      for (i=0; i<numFecTblEntries && strnicmp(SrvTbl[j].FecName,FecTbl[i].Name,FEC_NAME_SIZE); i++);
      /* i< numFecTblEntries means FEC entry found. */
      if (srvidx != NULL) *srvidx = j;	
      /* Jump out if FEC entry is complete (has protocol information) */
      if (i < numFecTblEntries && FecTbl[i].TransportProtocol) ccerr(0); /* returned cached entry */
    }
    if (isServiceRequest) 
    { /* Asking the "NETWORK" for the address */
      if (numSrvTblEntries == 0 && j == 0) 
      { /* I don't have a configured ENS (GENS)! Leave room for them, because I'm probably trying to find it */
        numFecTblEntries++; numSrvTblEntries++; 
      }
      if (numFecTblEntries >= FecTblSize) ccerr(out_of_client_memory);
      if (numSrvTblEntries >= SrvTblSize) ccerr(out_of_client_memory);
      /* "NETWORK" is not yet in the cached tables so add it */
      i = numFecTblEntries;
      strncpy(FecTbl[i].Name,SRVFEC_NAME,FEC_NAME_SIZE);
      strncpy(FecTbl[i].IP,gtNetSrvAddr,ADDR_SIZE);
      ip = inet_addr(gtNetSrvAddr);
      memcpy(&tNetSrvAddr,&ip,sizeof(struct in_addr));
      memcpy(FecTbl[i].IPh_addr,&tNetSrvAddr.s_addr,sizeof(struct in_addr));
      i = numSrvTblEntries;	
      strncpy(SrvTbl[i].FecName,SRVFEC_NAME,FEC_NAME_SIZE);
      strncpy(SrvTbl[i].ExportName,SRVEXP_NAME,EXPORT_NAME_SIZE);
      strncpy(SrvTbl[i].EqmName,SRVEQM_NAME,EQM_NAME_SHORTSIZE);
      if (srvidx != NULL) *srvidx = i;
      numFecTblEntries++; /* add "NETWORK" */
      numSrvTblEntries++; 
      feclog("NETWORK: added network services to the FEC table");
      if (tineDebug) dbglog("using network services as FEC address");
      ccerr(0); /* Jump out & return 0 */
    } 
    /* Buffer context */
    if (context != NULL) strncpy(cntxt,context,CONTEXT_NAME_SIZE);      
    /* Replace context with "#2" if empty. Legacy ENS request (#2 => return both FEC and EQM addresses) */
    if (strlen(cntxt) == 0) sprintf(cntxt,"#2");
          
    cc = queryNameServerForAddress(cntxt,expName,&fec,&srv);

    /* Jump to err if not successful */
    if (cc) ccerr(cc);
    /* If no FecTbl in cache, create a new one */
    if (j == numSrvTblEntries)
    {
      if (numSrvTblEntries < SrvTblSize) 
      {
        numSrvTblEntries++;
        isNew = TRUE;
      }
      else ccerr(resources_exhausted);
    }
    /* Fill FecTbl entry with data */
    memcpy(&SrvTbl[j],&srv,sizeof(ExpDataStruct));
    if (context != NULL && strnicmp(context,SrvTbl[j].EqmContext,CONTEXT_NAME_SIZE) != 0)
    {
      /* context was apparently decorated */
      strncpy(SrvTbl[j].EqmContext,context,CONTEXT_NAME_SIZE);
    }
    if (srvidx != NULL) *srvidx = j;
    /* Look for FEC name in FecTbl[] */
    for (i=0; i<numFecTblEntries && strnicmp(SrvTbl[j].FecName,FecTbl[i].Name,FEC_NAME_SIZE); i++);
    /* No entry found => create a new one. */
    if (i == numFecTblEntries)
    {
      if (numFecTblEntries < FecTblSize) 
      {
        numFecTblEntries++;
        isNew = TRUE;
      }
      else ccerr(buffer_too_small);
    }      
    /* i must be > 0! Do not under any circumstances overwrite the ENS or GENS addresses! */
    if (i > 1) memcpy(&FecTbl[i],&fec,sizeof(FecDataStruct));
    strncpy(fecName,FecTbl[i].Name,FEC_NAME_SIZE);
    if (isNew) 
    {
      feclog("ENS: %s [FEC %s, EQM %s] added",SrvTbl[j].ExportName,SrvTbl[j].FecName,SrvTbl[j].EqmName);
      msglog(0,"ENS: %s [FEC %s, EQM %s] added",SrvTbl[j].ExportName,SrvTbl[j].FecName,SrvTbl[j].EqmName);
#     ifndef FS_RDONLY
      if (!isAddressInCacheFile(gAddrCacheFilePath,j,i)) addAddressToCacheFile(gAddrCacheFilePath,j,i);
#     endif
    }
err:
    /* Jump back */
    if ((cc == link_not_open || cc == connection_timeout) && retry-- > 0) 
      goto retryRequest;  
    /* Jump out */
    if (cc == 0) break;                      /* finished */
    feclog("ENS: /%s/%s [%s] : %s",
      context != NULL ? context : "",
	    expName != NULL ? expName : "",
      fecName != NULL ? fecName : "",erlst[cc]);
    ToggleNameServer();
  }
  /* i is the found/new index in FecTbl[] */
  return cc ? -cc : i;                      /* return FEC index */
}
#ifndef FS_RDONLY
char gAddrCacheFilePath[128];
void establishAddrCacheFilePath(void)
{
  static int done = FALSE;
# ifdef FS_DELIMITER
  char *dptr,fpath[128];
  if (done) return;
# if defined(WIN32) || defined(DOS)
  if ((dptr=getenv("SystemDrive")) == NULL) dptr = "C:";
# elif defined(UNIX) || defined(VXWORKS) || defined(MACOS)
  dptr = "/var/tmp";
# else
  dptr = ".";
# endif
  sprintf(fpath,"%s%ctine%ccache%c",dptr,FS_DELIMITER,FS_DELIMITER,FS_DELIMITER);
  if (makeDirectoryPath(fpath)) strncpy(gAddrCacheFilePath,fpath,128);
  else strncpy(gAddrCacheFilePath,dptr,128);
# endif
  done = TRUE;
  return;
}
int isInSrvFile(char *filepath,char *ctxName,char *expName,char *eqmName,char *fecName)
{
  FILE *fp = NULL;
  int cc = 0,done = 0,found = FALSE;
  int nam_col=-1,fec_col=-1,eqm_col=-1,ctx_col=-1;
  char filename[256];
  char str[256],hdr[256];

  if (filepath == NULL) ccerr(argument_list_error);
  if (ctxName == NULL || expName == NULL || eqmName == NULL || fecName == NULL) ccerr(argument_list_error);
  sprintf(filename,"%s%s",filepath,EQPDB_FILE);
  if ((fp=fopen(filename,"rt")) == NULL) ccerr(no_such_file);
  for (done=0; fgets(str,255,fp) != NULL;)   /* scan file       */
  {
    strtrm(str);
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((nam_col=findcol(hdr,"NAME")) < 0) ccerr(no_such_column);
      if ((fec_col=findcol(hdr,"FEC_NAME")) < 0) ccerr(no_such_column);
      if ((eqm_col=findcol(hdr,"EQP_MODULE")) < 0) ccerr(no_such_column);
      if ((ctx_col=findcol(hdr,"CONTEXT")) < 0) ccerr(no_such_column);
      done = 1;
      continue;
    }
    if (namcmp(expName,str,nam_col,EXPORT_NAME_SIZE)) continue;
    if (namcmp(ctxName,str,ctx_col,CONTEXT_NAME_SIZE)) continue;
    if (namcmp(eqmName,str,eqm_col,EQM_NAME_SHORTSIZE)) continue;
    if (namcmp(fecName,str,fec_col,FEC_NAME_SIZE)) continue;
    found = TRUE;
    break;
  }
  fclose(fp); fp = NULL;
err:
  return found;
}
int isInFecFile(char *filepath,char *fecName,char *ipAddr,int portOffset,int tineProtocol)
{
  FILE *fp = NULL;
  int cc = 0,poff = -1,prot = 5,done = 0,found = FALSE;
  int fec_col=-1,ip_col=-1,port_col=-1,prot_col=-1;
  char filename[256];
  char str[256],hdr[256],scratch[64];

  if (filepath == NULL) ccerr(argument_list_error);
  if (fecName == NULL) ccerr(argument_list_error);
  sprintf(filename,"%s%s",filepath,FECDB_FILE);
  if ((fp=fopen(filename,"r")) == NULL) ccerr(no_such_file);
  for (done=0; fgets(str,255,fp) != NULL;)   /* scan file       */
  {
    strtrm(str);
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((fec_col=findcol(hdr,"FEC_NAME")) < 0) ccerr(no_such_column);
      if ((ip_col=findcol(hdr,"IP_ADDR")) < 0) ccerr(no_such_column);
      if ((port_col=findcol(hdr,"PORT_OFFSET")) < 0) ccerr(no_such_column);
      if ((prot_col=findcol(hdr,"TINE_PROTOCOL")) < 0) break;
      done = 1;
      continue;
    }
    if (namcmp(fecName,str,fec_col,FEC_NAME_SIZE)) continue;
    if (namcmp(ipAddr,str,ip_col,ADDR_SIZE)) continue;
    getColumnStr(port_col,str,scratch,ADDR_SIZE);
    poff = atoi(scratch);
    if (poff != portOffset) continue;
    getColumnStr(prot_col,str,scratch,ADDR_SIZE);
    prot = atoi(scratch);
    if (prot != tineProtocol) continue;
    found = TRUE;
    break;
  }
err:
  if (fp) fclose(fp);
  return found;
}
int isAddressInCacheFile(char *filepath,int srvidx,int fecidx)
{
  if (filepath == NULL) return FALSE;
  /* check names and addresses from the local cache entries for srvidx and fecidx ! */
  if (srvidx < 0 || srvidx >= numSrvTblEntries) return FALSE;
  if (fecidx < 0 || fecidx >= numFecTblEntries) return FALSE;
  if (!isInSrvFile(filepath,SrvTbl[srvidx].EqmContext,SrvTbl[srvidx].ExportName,SrvTbl[srvidx].EqmName,SrvTbl[srvidx].FecName))
    return FALSE;
  if (!isInFecFile(filepath,FecTbl[fecidx].Name,FecTbl[fecidx].IP,FecTbl[fecidx].PortOffset,FecTbl[fecidx].TineProtocol))
    return FALSE;
  /* if we made it this far, it's in the cache ! */
  return TRUE;
}
static char fecdbhdr[] = "FecName,FecNetwork,FecNode,IPaddr,PortOffset,TineProtocol\n";
static char eqpdbhdr[] = "Name,FecName,EqpModule,Context,SubSystem\n";
int addAddressToCacheFile(char *filepath,int srvidx,int fecidx)
{
  FILE *fp = NULL, *fpb = NULL;
  int cc = 0,done = 0,found = FALSE;
  int nam_col=-1,fec_col=-1,eqm_col=-1,ctx_col=-1,ip_col=-1,port_col=-1,prot_col=-1;
  char filename[256],filetemp[256];
  char str[256],hdr[256];
  char *expName,*ctxName,*eqmName,*fecName;

  if (filepath == NULL) ccerr(argument_list_error);
  if (srvidx < 0 || srvidx >= numSrvTblEntries) ccerr(invalid_index);
  if (fecidx < 0 || fecidx >= numFecTblEntries) ccerr(invalid_index);
  ctxName = SrvTbl[srvidx].EqmContext;
  expName = SrvTbl[srvidx].ExportName;
  eqmName = SrvTbl[srvidx].EqmName;
  fecName = SrvTbl[srvidx].FecName;
  sprintf(filetemp,"%s%s",filepath,EQPDB_FILE_TEMP);
  if ((fpb=fopen(filetemp,"wt")) == NULL) ccerr(file_error);
  sprintf(filename,"%s%s",filepath,EQPDB_FILE);
  if ((fp=fopen(filename,"rt")) == NULL) 
  { /* no cache file found ! */
    fprintf(fpb,eqpdbhdr);
    goto appendSrvDbFile;
  }
  for (done=0; fgets(str,255,fp) != NULL;)   /* scan file       */
  {
    strtrm(str);
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((nam_col=findcol(hdr,"NAME")) < 0) ccerr(no_such_column);
      if ((fec_col=findcol(hdr,"FEC_NAME")) < 0) ccerr(no_such_column);
      if ((eqm_col=findcol(hdr,"EQP_MODULE")) < 0) ccerr(no_such_column);
      if ((ctx_col=findcol(hdr,"CONTEXT")) < 0) ccerr(no_such_column);
      done = 1;
      fprintf(fpb,hdr); fprintf(fpb,"\n");
      continue;
    }
    if (!namcmp(expName,str,nam_col,EXPORT_NAME_SIZE) && 
        !namcmp(ctxName,str,ctx_col,CONTEXT_NAME_SIZE)) 
    { /* found the reference, now update the particulars ... */
      found = TRUE;
      fprintf(fpb,"%.16s,%.16s,%.8s,%.16s,%.16s\n",
        SrvTbl[srvidx].ExportName,SrvTbl[srvidx].FecName,SrvTbl[srvidx].EqmName,
        SrvTbl[srvidx].EqmContext,SrvTbl[srvidx].SubSystem); 
    }
    else
    { /* that wasn't it, spit the line back out */
      fprintf(fpb,str); fprintf(fpb,"\n");
    }
  }
  fclose(fp); fp = NULL;
appendSrvDbFile:
  if (!found)
  { /* not found : add to file cache */
    fprintf(fpb,"%.16s,%.16s,%.8s,%.16s,%.16s\n",
      SrvTbl[srvidx].ExportName,SrvTbl[srvidx].FecName,SrvTbl[srvidx].EqmName,
      SrvTbl[srvidx].EqmContext,SrvTbl[srvidx].SubSystem); 
  }
  fclose(fpb); fpb = NULL;
  unlink(filename);
  rename(filetemp,filename);

  found = FALSE;
  sprintf(filetemp,"%s%s",filepath,FECDB_FILE_TEMP);
  if ((fpb=fopen(filetemp,"wt")) == NULL) ccerr(file_error);
  sprintf(filename,"%s%s",filepath,FECDB_FILE);
  if ((fp=fopen(filename,"rt")) == NULL) 
  { /* no cache file found ! */
    fprintf(fpb,fecdbhdr);
    goto appendFecDbFile;
  }
  for (done=0; fgets(str,255,fp) != NULL;)   /* scan file       */
  {
    strtrm(str);
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((fec_col=findcol(hdr,"FEC_NAME")) < 0) ccerr(no_such_column);
      if ((ip_col=findcol(hdr,"IP_ADDR")) < 0) ccerr(no_such_column);
      if ((port_col=findcol(hdr,"PORT_OFFSET")) < 0) ccerr(no_such_column);
      prot_col = findcol(hdr,"TINE_PROTOCOL");
      done = 1;
      fprintf(fpb,fecdbhdr);
      continue;
    }
    if (!namcmp(fecName,str,fec_col,FEC_NAME_SIZE)) 
    { /* found it, so update it ! */
      found = TRUE;
      fprintf(fpb,"%s,%02x%02x%02x%02x,%02x%02x%02x%02x%02x%02x,%.16s,%d,%d\n",
            FecTbl[fecidx].Name,
            FecTbl[fecidx].IPXNetwork[0],FecTbl[fecidx].IPXNetwork[1],FecTbl[fecidx].IPXNetwork[2],FecTbl[fecidx].IPXNetwork[3],
            FecTbl[fecidx].IPXNode[0],FecTbl[fecidx].IPXNode[1],FecTbl[fecidx].IPXNode[2],
            FecTbl[fecidx].IPXNode[3],FecTbl[fecidx].IPXNode[4],FecTbl[fecidx].IPXNode[5],
            FecTbl[fecidx].IP,FecTbl[fecidx].PortOffset,FecTbl[fecidx].TineProtocol);
    }
    else
    { /* that wasn't it, so spit the line back out */
      fprintf(fpb,str);
      if (prot_col < 0) fprintf(fpb,",5"); /* worst case */
      fprintf(fpb,"\n");
    }
  }
  if (fp) fclose(fp); fp = NULL;
appendFecDbFile:
  if (!found)
  { /* not found : add to file cache */
    fprintf(fpb,"%s,%02x%02x%02x%02x,%02x%02x%02x%02x%02x%02x,%.16s,%d,%d\n",
          FecTbl[fecidx].Name,
          FecTbl[fecidx].IPXNetwork[0],FecTbl[fecidx].IPXNetwork[1],FecTbl[fecidx].IPXNetwork[2],FecTbl[fecidx].IPXNetwork[3],
          FecTbl[fecidx].IPXNode[0],FecTbl[fecidx].IPXNode[1],FecTbl[fecidx].IPXNode[2],
          FecTbl[fecidx].IPXNode[3],FecTbl[fecidx].IPXNode[4],FecTbl[fecidx].IPXNode[5],
          FecTbl[fecidx].IP,FecTbl[fecidx].PortOffset,FecTbl[fecidx].TineProtocol);
  }
  fclose(fpb); fpb = NULL;
  unlink(filename);
  rename(filetemp,filename);

err:
  if (fp) fclose(fp);
  if (fpb) fclose(fpb);
  return cc;
}
#endif
/**
 * \brief Get Data and FEC data from FEC and RPC-databases, or only FEC data.
 * 
 *  <ul>
 *  <li>Function 1. Find an entry in the local RPC Database file which corresponds to
 *  a given export name. Overwrite context and fecName. Store this entry in
 *  SrvTbl and write its index to srvidx. Then Function 2.</li>
 *  <li>Function 2. Find an entry in the local FEC Database file which corresponds to
 *  a given fec name.</li>
 *  </ul>
 *  
 * If expName is set to NULL, the function is used only as Function 2. 
 * In this case fecName is an input parameter. otherwise it is used as 
 * Function 1 and Function 2, with fecName as an output parameter.
 * 
 * \param context Function 1: context (output). Function 2: Not used.  
 * \param expName Decides if this function 1 and 2 or only function 2 is used. Function 1: export name. Function 2: NULL.
 * \param fecName Function 1: Fec name (output). Function 2: FEC name (input).
 * \param srvidx Function 1: Index of found entry in SrvTbl. Function 2: Not used.
 * 
 * \see gSrvDataBaseFullName
 * \see gFecDataBaseFullName
 * \see SrvTbl
 * \see FecTbl
 * .
 * \todo Remove gotos and jumps.
 */
int GetAddressFromNameFile(char *filepath,char *context,char *expName,char *fecName,int *srvidx)
{
  FILE *fp = NULL;
  DWORD haddr;
  int i = -1,cc = 0,done = 0,nam_col=-1,fec_col=-1,eqm_col=-1,cxt_col=-1,csv_col=-1;
  int chkAmbiguous = FALSE, hasIncremented = FALSE;
  char filename[256];
  char str[256],hdr[256],scratch[64], notedContext[CONTEXT_NAME_SIZE+4], *contxt = "";

  if (filepath == NULL) ccerr(argument_list_error);
  if (srvidx != NULL) *srvidx = -1;
  if (fecName == NULL) ccerr(non_existent_fec);
  if (expName == NULL) goto acquireFecName;     /* only interested in FEC address*/
  if (strlen(expName) == 0) ccerr(non_existent_elem);
  if (context != NULL && stricmp(context,"DEFAULT")) contxt = context;
  sprintf(filename,"%s%s",filepath,EQPDB_FILE);
  if ((fp=fopen(filename,"rt")) == NULL) ccerr(no_such_file);
  for (done=0; fgets(str,255,fp) != NULL;)   /* scan file       */
  {
    strtrm(str);
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
       
    /* Read header. */
    /* done means header has been read */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((nam_col=findcol(hdr,"NAME")) < 0) ccerr(no_such_column);
      if ((fec_col=findcol(hdr,"FEC_NAME")) < 0) ccerr(no_such_column);
      if ((eqm_col=findcol(hdr,"EQP_MODULE")) < 0) ccerr(no_such_column);
      if ((cxt_col=findcol(hdr,"CONTEXT")) < 0) ccerr(no_such_column);
      done = 1;
      continue;
    }
    if (!namcmp(expName,str,nam_col,EXPORT_NAME_SIZE))            /* found entry in file */
    {
      if (chkAmbiguous)
      {
        getColumnStr(cxt_col,str,scratch,CONTEXT_NAME_SIZE);
        if (strnicmp(notedContext,scratch,CONTEXT_NAME_SIZE))
        { /* same server different context */
          getColumnStr(fec_col,str,scratch,FEC_NAME_SIZE);
          if (strnicmp(SrvTbl[i].FecName,scratch,FEC_NAME_SIZE))
          { /* pointing to different FECs */
            if (hasIncremented) numSrvTblEntries--;
            ccerr(ambiguous);
          }
        }
        continue;
      }
      for (i=0; i<numSrvTblEntries; i++)
      {
        if (strnicmp(expName,SrvTbl[i].ExportName,EXPORT_NAME_SIZE)) continue;
        if (!strncmp(expName,GLOBALSRV,EXPORT_NAME_SIZE) && strlen(contxt) == 0) break;
        if (strnicmp(contxt,SrvTbl[i].EqmContext,CONTEXT_NAME_SIZE)) continue;
        break;
      }
      if (i == numSrvTblEntries)
      {
        hasIncremented = TRUE;
        if (numSrvTblEntries < SrvTblSize) numSrvTblEntries++;
        else ccerr(buffer_too_small);
      }
      getColumnStr(nam_col,str,SrvTbl[i].ExportName,EXPORT_NAME_SIZE);
      getColumnStr(fec_col,str,SrvTbl[i].FecName,FEC_NAME_SIZE);
      getColumnStr(eqm_col,str,SrvTbl[i].EqmName,EQM_NAME_SHORTSIZE);
      getColumnStr(cxt_col,str,scratch,CONTEXT_NAME_SIZE);
      if (strlen(contxt) > 0)
      {
        if (strnicmp(contxt,scratch,CONTEXT_NAME_SIZE)) 
        {
          if (hasIncremented) numSrvTblEntries--;
          i = -1; /* not a match */
          continue;
        }
        strncpy(SrvTbl[i].EqmContext,scratch,CONTEXT_NAME_SIZE);
      }
      else 
      {
        strncpy(notedContext,scratch,CONTEXT_NAME_SIZE);
        SrvTbl[i].EqmContext[0] = 0;
        chkAmbiguous = TRUE;
      }
      feclog("EFL: %s [FEC %s, EQM %s] added",
	      SrvTbl[i].ExportName,SrvTbl[i].FecName,SrvTbl[i].EqmName);
      if (srvidx != NULL) *srvidx = i;
      if (!chkAmbiguous) break;  /* finished if context given */
    }
  }
  if (i < 0) ccerr(non_existent);
  strncpy(fecName,SrvTbl[i].FecName,FEC_NAME_SIZE);
  fclose(fp); fp = NULL;

acquireFecName:
  if (strlen(fecName) == 0) ccerr(non_existent_fec);
  sprintf(filename,"%s%s",filepath,FECDB_FILE);
  if ((fp=fopen(filename,"r")) == NULL) ccerr(no_such_file);
  for (done=0; fgets(str,255,fp) != NULL;)   /* scan file       */
  {
    strtrm(str);
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((csv_col=findcol(hdr,"FEC_NAME")) < 0) ccerr(no_such_column);
      done = 1;
      continue;
    }
    if (!namcmp(fecName,str,csv_col,FEC_NAME_SIZE))           /* found entry in file */
    {
      for (i=0; i<numFecTblEntries && strnicmp(fecName,FecTbl[i].Name,FEC_NAME_SIZE); i++);
      if (i == numFecTblEntries)
      {
        if (numFecTblEntries < FecTblSize) numFecTblEntries++;
        else ccerr(buffer_too_small);
      }
      getColumnStr(csv_col,str,FecTbl[i].Name,FEC_NAME_SIZE);
      if ((csv_col=findcol(hdr,"FEC_NETWORK")) < 0) ccerr(no_such_column);
      getColumnStr(csv_col,str,scratch,16);
      HexToBytes(scratch,FecTbl[i].IPXNetwork,8);
      if ((csv_col=findcol(hdr,"FEC_NODE")) < 0) ccerr(no_such_column);
      getColumnStr(csv_col,str,scratch,16);
      HexToBytes(scratch,FecTbl[i].IPXNode,12);
      if ((csv_col=findcol(hdr,"IP_ADDR")) < 0) ccerr(no_such_column);
      getColumnStr(csv_col,str,FecTbl[i].IP,ADDR_SIZE);
      haddr = inet_addr(FecTbl[i].IP);
      memcpy(FecTbl[i].IPh_addr,&haddr,4);
      feclog("EFL: FEC %s [%s %s] added",FecTbl[i].Name,scratch,FecTbl[i].IP);
      if ((csv_col=findcol(hdr,"PORT_OFFSET")) >= 0)
      {
        getColumnStr(csv_col,str,scratch,ADDR_SIZE);
        FecTbl[i].PortOffset = atoi(scratch);
      }
      else FecTbl[i].PortOffset = 0;
      if ((csv_col=findcol(hdr,"TINE_PROTOCOL")) >= 0)
      {
        getColumnStr(csv_col,str,scratch,ADDR_SIZE);
        FecTbl[i].TineProtocol = atoi(scratch);
      }
      else if ((csv_col=findcol(hdr,"VERSION")) >= 0)
      {
        getColumnStr(csv_col,str,scratch,ADDR_SIZE);
        FecTbl[i].TineProtocol = atoi(scratch) == 4 ? 6 : 5;
      }
      else FecTbl[i].TineProtocol = 0;
      break;
    }
  }
  if (i < 0) cc = non_existent;
err:
  if (fp) fclose(fp);
  if (cc && cc != no_such_file)
  {
    if (expName == NULL) expName = "(null)";
    feclog("EFL: /%s/%s : %s",contxt,expName,erlst[cc]);
  }
  return cc ? -cc : i;
}

int dbaseloaded = 0;
int MaxTblAllocationSize = -1;
int reserveFecAddressSpace(void)
{
  int cc = 0;
  if (SrvTbl != NULL && FecTbl != NULL) 
  {
    /* already reserved */
    return 0;
  }
  if (SrvTblSize == 0) SrvTblSize = MaxNumConnections;
  if (FecTblSize == 0) FecTblSize = MaxNumConnections;
  if (MaxTblAllocationSize > 0) SrvTblSize = FecTblSize = MaxTblAllocationSize;

  if ((SrvTbl == NULL) && /* previously allocated ? */
     !(SrvTbl=(ExpDataStruct *)SystemCalloc(SrvTblSize,sizeof(ExpDataStruct))))
    ccerr(out_of_client_memory);

  if ((FecTbl == NULL) && /* previously allocated ? */
      !(FecTbl=(FecDataStruct *)SystemCalloc(FecTblSize,sizeof(FecDataStruct))))
    ccerr(out_of_client_memory);

  feclog("Reserved %d client connection link entries",SrvTblSize);
  feclog("Reserved %d client connection addr entries",FecTblSize);

err:
  return cc;
}
/**
 * \internal 
 * \brief Initializes different things relating to name resolution.
 * 
 * <ul>
 * <li>Initializes tineHomePath,gSrvDataBaseFullName,gFecDataBaseFullName</li>
 * <li>Calls GetAllocSizesFromNameServer, initializes FECAllocNum, SrvTblSize</li>
 * <li>Allocates SrvTbl and FecTbl</li>
 * <li>Initialize conTbl with name server entry.</li>
 * </ul>
 * 
 * \see tineHomePath,gSrvDataBaseFullName,gFecDataBaseFullName
 * \see dbaseloaded
 */
int initFecTable(void)
{
  char *ptr;
  int i,cc = 0;

  if (dbaseloaded) return 0;

  /* Determine gSrvDataBaseFullName and gFecDataBaseFullName from environment variables. */ 
  tineHomePath[0] = 0;
  if (gIsEquipmentNameServer)
  {
    feclog("Server is the ENS: ignore environment settings");
  }
  else if ((ptr=getenv("TINE_HOME")) != NULL || (ptr=getenv("CONTROLDB")) != NULL)
  {
    strncpy(tineHomePath,ptr,80);
#   ifdef FS_DELIMITER
    if (((i=(int)strlen(tineHomePath)) > 0) &&
        (tineHomePath[i-1] != FS_DELIMITER))  /* no delimiter ! */
    {
      tineHomePath[i] = FS_DELIMITER; tineHomePath[i+1] = 0;
    }
#   endif
  }
  if ((ptr=getenv("USE_LOOPBACK")) != NULL)
  {
    if (!stricmp(ptr,"TRUE")) gUseLoopback = TRUE;
  }
  if ((cc=reserveFecAddressSpace()) != 0) ccerr(cc);
  numSrvTblEntries = numFecTblEntries = 0;
  addNameServerToTables();
# ifndef FS_RDONLY
  establishAddrCacheFilePath();
# endif
  dbaseloaded = 1;
err:
  if (cc) nConnectionTableEntries = 0;
  return cc;
}

/**
 * \internal 
 * \brief Calls GetNameServerAddress() and Initializes NameServerData and 
 * connNameServer. Sets both output parameters to the MaxNumConnections.
 * 
 * Uses the global variable NameServerLoaded as a lock which ensures
 * that this function is called only once.
 * 
 * \see MaxNumConnections
 * \see NameServerData
 * \see connNameServer
 * \see GetNameServerAddress()
 * 
 */
int addNameServerToTables(void)
{
  int cc=0;

  InitSystemMutexSet();
  if (WaitForSystemMutex(hLinkTblMutex,-1) != 0) ccerr(semaphore_error);
  if (!NameServerLoaded)
  {
    if ((cc=GetNameServerAddress()) != 0) ccerr(cc);
    strncpy(NameServerData.FecName,NameServer[0].Name,FEC_NAME_SIZE);
    strncpy(NameServerData.ExportName,CSHOSTS,EXPORT_NAME_SIZE);
    strncpy(NameServerData.EqmName,CSHOSTS,EQM_NAME_SHORTSIZE);
    strncpy(connNameServer.expName,CSHOSTS,EXPORT_NAME_SIZE);
    strncpy(connNameServer.devName,"#0",DEVICE_NAME_SIZE);
    strncpy(connNameServer.prpName,"SIZES",PROPERTY_NAME_SIZE);
    connNameServer.formatOut = CF_SHORT;
    connNameServer.formatIn = CF_NULL;
    connNameServer.srvIdx = 0;
    connNameServer.sizeOut = 2;
    connNameServer.sizeIn = 0;
    connNameServer.tineProtocol = gTineServicesProtocol;
    connNameServer.mode = CM_REGISTER;
    connNameServer.blknumIn = 1;
    connNameServer.numblksIn = 1;
    connNameServer.nextDataPtr = 0;
    conTbl[0] = &connNameServer; /* Hardwire first call */
    nConnectionTableEntries = 1;
    if (SrvTbl != NULL || FecTbl != NULL)
    {
      memcpy(&SrvTbl[0],&NameServerData,sizeof(ExpDataStruct));
      memcpy(&FecTbl[0],&NameServer[0],sizeof(FecDataStruct));
      conTbl[0] = &connNameServer; /* Hardwire first call */
      /* add the GENS information as well ... */
      memcpy(&FecTbl[1],&NameServer[0],sizeof(FecDataStruct));
      strcpy(FecTbl[1].Name,"GENS");
      FecTbl[1].PortOffset += 101;
      memcpy(&SrvTbl[1],&NameServerData,sizeof(ExpDataStruct));
      strcpy(SrvTbl[1].FecName,"GENS");
      strcpy(SrvTbl[1].EqmName,"GRPEQM");
      strcpy(SrvTbl[1].ExportName,"GENS");
      numSrvTblEntries = numFecTblEntries = 2;
    }
    NameServerLoaded = 1;
  }
err:
  ReleaseSystemMutex(hLinkTblMutex);
  return cc;
}
